#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

# pylint: disable=consider-using-enumerate

# Imports from the python standard library
import os.path
import pickle

# Imports from third parties
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize as opt

from pyomo.environ import (
    ConcreteModel,
    Param,
    Set,
    Var,
    Reals,
    Objective,
    minimize,
    SolverFactory,
    Constraint,
    exp,
    log,
)

# Imports from IDAES namespace
from idaes.core.surrogate.pysmo.sampling import FeatureScaling as fs
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

__author__ = "Oluwamayowa Amusat"

"""
The purpose of this file is to perform radial basis functions in Pyomo.
"""


class FeatureScaling:
    """

    A class for scaling and unscaling input and output data. The class contains two main methods: ``data_scaling_minmax`` and ``data_unscaling_minmax``
    """

    def __init__(self):
        pass

    @staticmethod
    def data_scaling_minmax(data):
        """
        ``data_scaling_minmax`` performs column-wise min-max scaling on the input dataset.

        Args:
            data : The input data set to be scaled. Must be a numpy array or dataframe.

        Returns:
            (tuple): tuple containing:
                - **scaled_data**  : A 2-D Numpy Array containing the scaled data. All array values will be between [0, 1].
                - **data_minimum** : A 2-D row vector containing the column-wise minimums of the input data.
                - **data_maximum** : A 2-D row vector containing the column-wise maximums of the input data.

        Raises:
            TypeError: Raised when the input data is not a numpy array or dataframe

        """
        # Confirm that data type is an array or DataFrame
        if isinstance(data, np.ndarray):
            input_data = data
            data_headers = []
        elif isinstance(data, pd.DataFrame):
            input_data = data.values
            data_headers = data.columns.values.tolist()
        else:
            raise TypeError(
                "original_data_input: Pandas dataframe or numpy array required."
            )

        if input_data.ndim == 1:
            input_data = input_data.reshape(len(input_data), 1)
        data_minimum = np.min(input_data, axis=0)
        data_maximum = np.max(input_data, axis=0)
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        scaled_data = (input_data - data_minimum) / scale
        # scaled_data = (input_data - data_minimum)/(data_maximum - data_minimum)
        data_minimum = data_minimum.reshape(1, data_minimum.shape[0])
        data_maximum = data_maximum.reshape(1, data_maximum.shape[0])

        if len(data_headers) > 0:
            scaled_data = pd.DataFrame(scaled_data, columns=data_headers)
        return scaled_data, data_minimum, data_maximum

    @staticmethod
    def data_unscaling_minmax(x_scaled, x_min, x_max):
        """

        ``data_unscaling_minmax`` performs column-wise un-scaling on the a minmax-scaled input dataset.

        Args:
            x_scaled (NumPy Array)  : Data to be un-scaled. Data values should be between 0 and 1.
            x_min (NumPy vector)    : :math:`n \\times 1` vector containing the actual minimum value for each column. Must contain same number of elements as the number of columns in x_scaled.
            x_max (NumPy vector)    : :math:`n \\times 1` vector vector containing the actual minimum value for each column. Must contain same number of elements as the number of columns in x_scaled.

        Returns:
            NumPy Array : A 2-D numpy array containing the scaled data, :math:`x_{min} + x_{scaled} * (x_{max} - x_{min})`

        Raises:
            IndexError: Raised when the dimensions of the arrays are inconsistent.

        """
        # Check if it can be evaluated. Will return index error if dimensions are wrong
        if x_scaled.ndim == 1:  # Check if 1D, and convert to 2D if required.
            x_scaled = x_scaled.reshape(len(x_scaled), 1)
        if (x_scaled.shape[1] != x_min.size) or (x_scaled.shape[1] != x_max.size):
            raise IndexError("Dimensionality problems with data for un-scaling.")
        unscaled_data = x_min + x_scaled * (x_max - x_min)
        return unscaled_data

    # @staticmethod
    # def data_scaling_standardization(data):
    #     # Confirm that data type is an array or DataFrame
    #     if isinstance(data, np.ndarray):
    #         input_data = data
    #     elif isinstance(data, pd.DataFrame):
    #         input_data = data.values
    #     else:
    #         raise TypeError('original_data_input: Pandas dataframe or numpy array required.')
    #
    #     if input_data.ndim == 1:
    #         input_data = input_data.reshape(len(input_data), 1)
    #
    #     data_mean = np.mean(input_data, axis=0)
    #     data_stdev = np.std(input_data, axis=0)
    #     scaled_data = (input_data - data_mean) / data_stdev
    #     data_mean = data_mean.reshape(1, data_mean.shape[0])
    #     data_stdev = data_stdev.reshape(1, data_stdev.shape[0])
    #     return scaled_data, data_mean, data_stdev


class RadialBasisFunctions:
    r"""
    The RadialBasisFunctions class generates a radial basis function fitting for a training data set.

    The class must first be initialized by calling **RadialBasisFunctions**. Regression is then carried out by calling the method ``training``.

    For a given dataset with n features :math:`x_{1},\ldots,x_{n}`, RadialBasisFunctions is able to consider six types of basis transformations:
        - Linear ('linear')
        - Cubic ('cubic')
        - Gaussian ('gaussian')
        - Multiquadric ('mq')
        - Inverse multiquadric ('imq')
        - Thin-plate spline ('spline')

    ``training`` selects the best hyperparameters (regularization parameter :math:`\lambda` and shape parameter :math:`\sigma`, where necessary) by evaluating the leave-one-out cross-validation error for each (:math:`\lambda,\sigma`) pair.

    It should be noted that the all the training points are treated as centres for the RBF, resulting in a square system.

    **Example:**

    .. code-block:: python

         # Initialize the class
        >>> d = RadialBasisFunctions(training_data, basis_function='gaussian', solution_method='pyomo', regularization=True))
        >>> p = d.get_feature_vector()

        # Train RBF model and predict output for an test data x_test
        >>> d.training()
        >>> predictions = d.predict_output(x_test)

    Args:
        XY_data (Numpy Array or Pandas Dataframe)             : The dataset for RBF training. **XY_data** is expected to contain the features (X) and output (Y) data, with the output values (Y) in the last column.

    Further details about the optional inputs may be found under the ``__init__`` method.

    """

    def __init__(
        self,
        XY_data,
        basis_function=None,
        solution_method=None,
        regularization=None,
        fname=None,
        overwrite=False,
    ):
        r"""

        Initialization of **RadialBasisFunctions** class.

        Args:
            XY_data (Numpy Array or Pandas Dataframe): The dataset for RBF training. **XY_data** is expected to contain feature and output information, with the output values (y) in the last column.

        Keyword Args:
            basis_function(str): The basis function transformation to be applied to the training data. Two classes of basis transformations are available for selection:

                - Fixed basis transformations, which require no shape parameter :math:`\sigma` :

                    (a)  'cubic'      : Cubic basis transformation
                    (b) 'linear'      : Linear basis transformation
                    (c) 'spline'      : Thin-plate spline basis transformation

                - Parametric basis transformations which require a shape parameter:

                    (a) 'gaussian'    : Gaussian basis transformation (Default)
                    (b) 'mq'          : Multiquadric basis transformation
                    (c) 'imq'         : Inverse multiquadric basis transformation


            solution_method(str): The method to be used for solving the RBF least squares optimization problem. Three options are available:

                (a) 'algebraic'  : The explicit algebraic method solves the least squares problem using linear algebra.
                (b) 'BFGS'       : This approach solves the least squares problem using SciPy's BFGS algorithm.
                (c) 'pyomo'      : This option solves the optimization problem in Pyomo with IPOPT as solver. This is the default.

            regularization(bool): This option determines whether or not the regularization parameter :math:`\lambda` is considered during RBF fitting. Default setting is True.


        Returns:
            **self** object with the input information

        Raises:
            ValueError: The input dataset is of the wrong type (not a NumPy array or Pandas Dataframe)

            Exception:
                * **basis_function** entry is not valid.
            Exception:
                * **solution_method** is not 'algebraic', 'pyomo' or 'bfgs'.
            Exception:
                - :math:`\lambda` is not boolean.

        **Example:**

        .. code-block:: python

            # Specify the gaussian basis transformation
            >>> d = RadialBasisFunctions(XY_data, basis_function='gaussian')

        """
        if not isinstance(overwrite, bool):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("overwrite must be boolean.")
        self.overwrite = overwrite
        if fname is None:
            fname = "solution.pickle"
            self.filename = "solution.pickle"
        elif (
            not isinstance(fname, str)
            or os.path.splitext(fname)[-1].lower() != ".pickle"
        ):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                'fname must be a string with extension ".pickle". Please correct.'
            )
        if (
            os.path.exists(fname) and overwrite is True
        ):  # Explicit overwrite done by user
            print(
                "Warning:",
                fname,
                "already exists; previous file will be overwritten.\n",
            )
            self.filename = fname
        elif os.path.exists(fname) and overwrite is False:  # User is not overwriting
            self.filename = (
                os.path.splitext(fname)[0]
                + "_v_"
                + pd.Timestamp.today().strftime("%m-%d-%y_%H%M%S")
                + ".pickle"
            )
            print(
                "Warning:",
                fname,
                'already exists; results will be saved to "',
                self.filename,
                '".\n',
            )
            # self.filename = 'solution.pickle'
        elif os.path.exists(fname) is False:
            self.filename = fname

        # Check data types and shapes
        if isinstance(XY_data, pd.DataFrame):
            xy_data = XY_data.values
            self.x_data_columns = list(XY_data.columns)[:-1]
        elif isinstance(XY_data, np.ndarray):
            xy_data = XY_data
            self.x_data_columns = list(range(XY_data.shape[1] - 1))
        else:
            raise ValueError('Pandas dataframe or numpy array required for "XY_data".')

        self.x_data_unscaled = xy_data[:, :-1]
        self.y_data_unscaled = xy_data[:, -1].reshape(xy_data.shape[0], 1)
        xy_data_scaled, self.data_min, self.data_max = fs.data_scaling_minmax(XY_data)
        x_data_scaled = xy_data_scaled[:, :-1]
        y_data_scaled = xy_data_scaled[:, -1]
        self.x_data = x_data_scaled.reshape(self.x_data_unscaled.shape)
        self.y_data = y_data_scaled.reshape(self.y_data_unscaled.shape)
        self.centres = xy_data_scaled[:, :-1]

        if solution_method is None:
            solution_method = "algebraic"
            self.solution_method = solution_method
            print("Default parameter estimation method is used.")
        elif not isinstance(solution_method, str):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("Invalid solution method. Must be of type <str>.")
        elif (
            (solution_method.lower() == "algebraic")
            or (solution_method.lower() == "pyomo")
            or (solution_method.lower() == "bfgs")
        ):
            solution_method = solution_method.lower()
            self.solution_method = solution_method
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                'Invalid solution method entered. Select one of ALGEBRAIC (solution_method="algebraic") , L-BFGS (solution_method="bfgs") or Pyomo optimization (solution_method="pyomo") methods. '
            )
        print("\nParameter estimation method: ", self.solution_method)

        if basis_function is None:
            basis_function = "gaussian"
            self.basis_function = basis_function
            print("Gaussian basis function is used.")
        elif not isinstance(basis_function, str):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("Invalid basis_function. Must be of type <str>.")
        elif (
            (basis_function.lower() == "linear")
            or (basis_function.lower() == "cubic")
            or (basis_function.lower() == "gaussian")
            or (basis_function.lower() == "mq")
            or (basis_function.lower() == "imq")
            or (basis_function.lower() == "spline")
        ):
            basis_function = basis_function.lower()
            self.basis_function = basis_function
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Invalid basis function entered. See manual for available options. "
            )
        print("Basis function: ", self.basis_function)

        if regularization is None:
            regularization = True
            self.regularization = regularization
        elif not isinstance(regularization, bool):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("Invalid basis_function. Must be boolean")
        elif (regularization is True) or (regularization is False):
            self.regularization = regularization
        print("Regularization done: ", self.regularization)

        # Results
        self.weights = None
        self.sigma = None
        self.regularization_parameter = None
        self.rmse = None
        self.output_predictions = None
        self.condition_number = None
        self.R2 = None
        self.x_data_min = None
        self.x_data_max = None
        self.y_data_min = None
        self.y_data_max = None
        self.solution_status = None

    def r2_distance(self, c):
        """
        The function r2_distance calculates Euclidean distance from the point or array c.

        """
        dist = self.x_data - c
        l2_distance = np.sqrt(np.sum((dist**2), axis=1))
        return l2_distance

    @staticmethod
    def gaussian_basis_transformation(x, shape_parameter):
        """
        The function gaussian_basis_transformation returns the element-by-element Gaussian transformation of the input data x.

        Args:
            x(NumPy Array): Input data to be transformed
            shape parameter(float): Shape parameter of the Gaussian function

        Returns:
            x_mod(NumPy Array): Gaussian transformation of the input data x

        Examples:
            Gaussian of numbers from 0 to 2 for a shape parameter of 2:
                [In]>>  rbf.RadialBasisFunctions.gaussian_basis_transformation(np.arange(3), 2)
                [Out]>> array([1.00000000e+00, 1.83156389e-02, 1.12535175e-07])

        For more information, see Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294

        """
        x_mod = np.exp(-1 * ((x * shape_parameter) ** 2))
        return x_mod

    @staticmethod
    def linear_transformation(x):
        """
        The function linear_transformation returns the element-by-element linear transformation of the input data x.

        Args:
            x(NumPy Array): Input data to be transformed

        Returns:
            x_mod(NumPy Array): Linear transformation of the input data x, x_mod = x

        Examples:
            Linear transformation of 0, 1 and 2:
                [In]>>  rbf.RadialBasisFunctions.linear_transformation(np.arange(3))
                [Out]>> array([0, 1, 2], dtype=int32)

        For more information, see Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294

        """
        x_mod = x**1
        return x_mod

    @staticmethod
    def cubic_transformation(x):
        """
        The function cubic_transformation returns the element-by-element cubic transformation of the input data x.

        Args:
            x(NumPy Array): Input data to be transformed

        Returns:
            x_mod(NumPy Array): Cubic transformation of the input data x, x_mod = (x ** 3)

        Examples:
            Cubic transformation of 0, 1 and 2:
                [In]>>  rbf.RadialBasisFunctions.cubic_transformation(np.arange(3))
                [Out]>> array([0, 1, 8], dtype=int32)

        For more information, see Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294
        """
        x_mod = x**3
        return x_mod

    @staticmethod
    def multiquadric_basis_transformation(x, shape_parameter):
        """
        The function multiquadric_basis_transformation returns the element-by-element Multi-quadric transformation of the input data x.

        Args:
            x(NumPy Array): Input data to be transformed
            shape parameter(float): Shape parameter of the Multiquadric function

        Returns:
            x_mod(NumPy Array): Multiquadric transformation of the input data x; x_mod = sqrt[(1 + (c.x)**2)] where c = shape parameter

        Examples:
            Multiquadric transformation of numbers from 0 to 2 for a shape parameter of 2:
                [In]>>  rbf.RadialBasisFunctions.multiquadric_basis_transformation(np.arange(3), 2)
                [Out]>> array([1.        , 2.23606798, 4.12310563])

        For more information, see
        (1) Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294

        (2) Santana-Quintero L.V., Montaño A.A., Coello C.A.C. (2010) A Review of Techniques for Handling Expensive Functions in Evolutionary Multi-Objective Optimization.
        In: Tenne Y., Goh CK. (eds) Computational Intelligence in Expensive Optimization Problems.
        https://link.springer.com/chapter/10.1007/978-3-642-10701-6_2

        """
        x_mod = np.sqrt(((x * shape_parameter) ** 2) + 1)
        # x_mod = np.sqrt(x**2 + shape_parameter**2)  # Alternative implementation
        return x_mod

    @staticmethod
    def inverse_multiquadric_basis_transformation(x, shape_parameter):
        """
        The function inverse_multiquadric_basis_transformation returns the element-by-element  inverse multiquadric transformation of the input data x.
        Direct inverse of the multiquadric basis transformation

        Args:
            x(NumPy Array): Input data to be transformed
            shape_parameter(float): Shape parameter of the inverse multiquadric function

        Returns:
            x_mod(NumPy Array): Inverse multiquadric transformation of the input data x; x_mod =  1 / sqrt[(1 + (c.x)**2)] where c = shape parameter

        Examples:
            Inverse multiquadric transformation of numbers from 0 to 2 for a shape parameter of 2:
                [In]>>  rbf.RadialBasisFunctions.inverse_multiquadric_basis_transformation(np.arange(3), 2)
                [Out]>>array([1.        , 0.4472136 , 0.24253563])

        For more information, see
        (1) Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294

        (2) Santana-Quintero L.V., Montaño A.A., Coello C.A.C. (2010) A Review of Techniques for Handling Expensive Functions in Evolutionary Multi-Objective Optimization.
        In: Tenne Y., Goh CK. (eds) Computational Intelligence in Expensive Optimization Problems.
        https://link.springer.com/chapter/10.1007/978-3-642-10701-6_2

        """
        x_mod = 1 / np.sqrt(((x * shape_parameter) ** 2) + 1)
        # x_mod = 1 / (np.sqrt(x ** 2 + shape_parameter ** 2))  # Alternative implementation
        return x_mod

    @staticmethod
    def thin_plate_spline_transformation(x):
        """
        The function thin_plate_spline_transformation returns the element-by-element spline transformation of the input data x.
        Direct inverse of the multiquadric basis transformation

        Args:
            x(NumPy Array): Input data to be transformed

        Returns:
            x_mod(NumPy Array): Spline transformation of the input data x; x_mod =  (x**2).ln(x)

        Except:
            RuntimeWarning: thrown up when ln(x)=0

        Examples:
            Inverse multiquadric transformation of numbers from 1 and 2:
                [In]>>  rbf.RadialBasisFunctions.thin_plate_spline_transformation(np.arange(1,3))
                [Out]>> array([0, 2.77258872])

        For more information, see
        (1) Hongbing Fang & Mark F. Horstemeyer (2006): Global response approximation with radial basis functions
        https://www.tandfonline.com/doi/full/10.1080/03052150500422294

        (2) Santana-Quintero L.V., Montaño A.A., Coello C.A.C. (2010) A Review of Techniques for Handling Expensive Functions in Evolutionary Multi-Objective Optimization.
        In: Tenne Y., Goh CK. (eds) Computational Intelligence in Expensive Optimization Problems.
        https://link.springer.com/chapter/10.1007/978-3-642-10701-6_2

        """
        # x_mod = (x ** 2) * np.log(x)
        # x_mod = np.nan_to_num(x_mod)
        with np.errstate(
            divide="ignore"
        ):  # catch division warnings in log function due to log(0)~=0
            log_x = np.log(x)
        with np.errstate(
            invalid="ignore"
        ):  # catch invalid warnings due to - Inf * 0 evaluations
            x_mod = (x**2) * log_x
        x_mod = np.nan_to_num(x_mod)
        return x_mod

    def basis_generation(self, r):
        """
        The function basis_generation converts the input data to the requisite basis specified by the user.
        This is done in two steps:

        1. The Euclidean distance from each of the points to each of the RBF centres is calculated by calling the r2_distance function.
        2. The distances evaluated in step 1 are transformed to the relevant basis selected by the user.

        Args:
            self(NumPy Array): contains, among other things, the input data
            r(float)        : The shape parameter required for the Gaussian, Multiquadric and Inverse multiquadric transformations.

        Returns:
            x_transformed(NumPy Array): Array of transformed data based on user-defined transformation function

        """

        basis_functions = np.zeros((self.x_data.shape[0], self.centres.shape[0]))
        for i in range(0, self.centres.shape[0]):
            basis_functions[:, i] = self.r2_distance(self.centres[i, :])

        # Initialization of x_transformed
        x_transformed = np.zeros((basis_functions.shape[0], basis_functions.shape[1]))

        if self.basis_function == "gaussian":
            x_transformed = self.gaussian_basis_transformation(basis_functions, r)
        elif self.basis_function == "linear":
            x_transformed = self.linear_transformation(basis_functions)
        elif self.basis_function == "cubic":
            x_transformed = self.cubic_transformation(basis_functions)
        elif self.basis_function == "mq":
            x_transformed = self.multiquadric_basis_transformation(basis_functions, r)
        elif self.basis_function == "imq":
            x_transformed = self.inverse_multiquadric_basis_transformation(
                basis_functions, r
            )
        elif self.basis_function == "spline":
            x_transformed = self.thin_plate_spline_transformation(basis_functions)
        return x_transformed

    @staticmethod
    def cost_function(theta, x, y):
        """
        This function is an implementation of the cost function for linear regression with BFGS:
                cost = [sum of square errors over m samples / (2 * m)]

        This is the objective function for the BFGS optimization problem.

        Args:
            theta        : polynomial coefficients/weights,  (n x 1) in size
            x            : array of features, (m x n) in size
            y            : actual output vector, size (m x 1)

        Returns:
            cost_value   : the cost value for the fit, the objective value of the optimization problem

        """
        y = y.reshape(y.shape[0], 1)
        y_prediction = np.matmul(x, theta)
        y_prediction = y_prediction.reshape(y_prediction.shape[0], 1)
        cost_value = (0.5 / x.shape[0]) * (np.sum((y - y_prediction) ** 2))
        return cost_value

    @staticmethod
    def gradient_function(theta, x, y):
        """
        This function is an implementation of the gradient function for linear regression:
                if
                    cost = [(A.x - y)^2 / 2m]
                then
                    gradient = [((A.x - y)* A) / m]

        This is the gradient function supplied to the BFGS optimization algorithm.

        Args:
            theta        : polynomial coefficients/weights,  (n x 1) in size
            x            : array of features, (m x n) in size
            y            : actual output vector, size (m x 1)

        Returns:
            grad_values   : the cost gradients for the fit, size (n x 1)

        """
        y = y.reshape(y.shape[0], 1)
        y_prediction = np.matmul(x, theta)
        y_prediction = y_prediction.reshape(y_prediction.shape[0], 1)
        t1 = (y_prediction - y) * x
        grad_values = (1 / x.shape[0]) * np.sum(t1, axis=0)
        grad_values = grad_values.reshape(
            theta.size,
        )
        return grad_values

    def bfgs_parameter_optimization(self, x, y):
        """
        This function performs parameter optimization using scipy's BFGS algorithm.
        It takes in the functions pre-defined functions cost_function and gradient_function as the cost and gradient functions.

        Args:
            x            : array of features, (m x n) in size
            y            : actual output vector, size (m x 1)

        Initialization:
            The initial weights are set to zero,
            init_phi = 0

        Returns:
            phi: The optimal linear regression weights found

        """
        init_phi = np.zeros((x.shape[1], 1))
        other_args = (x, y)
        phi = opt.fmin_bfgs(
            self.cost_function,
            init_phi,
            fprime=self.gradient_function,
            args=other_args,
            disp=False,
            gtol=1e-20,
        )
        return phi

    @staticmethod
    def explicit_linear_algebra_solution(x, y):
        """
        The function finds the explicit linear algebra solution to the reqularized problem (X+yI).A = B

        If:
            (x + yI).A = B,

        Then:
            A = inv(X + yI) * B

        where y is the regularization parameter and I is the identity matrix.

        Numpy's inverse and pseudoinverse functions have been used to calculate the inverse here.

        Args:
            x            : regularized array of features (x + yI), (m x n) in size
            y            : actual output vector, size (m x 1)

        Returns:
            phi          : optimal linear regression weights A

        For more details, see to Forrester et al.'s book "Engineering Design via Surrogate Modelling: A Practical Guide", https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470770801

        """
        # Find matrix inverse. Use pseudo-inverse if inverse is not available
        try:
            inverse_x = np.linalg.inv(x)
        except np.linalg.LinAlgError:
            inverse_x = np.linalg.pinv(x)

        phi = np.matmul(inverse_x, y)
        return phi

    @staticmethod
    def pyomo_optimization(x, y):
        """
        Pyomo implementation of least squares optimization problem:

                        Minimize cost = (y' - y) ^ 2
                        subject to: y' = Ax

        The problem is solved within Pyomo's framework using IPOPT as solver.

        Args:
            x            : array of features, (m x n) in size
            y            : actual output vector, size (m x 1)

        Returns:
            phi: The optimal linear regression weights found

        """
        model = ConcreteModel()

        pd.set_option("display.precision", 64)
        x_data = pd.DataFrame(x)
        y_data = pd.DataFrame(y)

        model.M = Set(initialize=x_data.index.values)  # Rows indices passed into set
        model.N = Set(
            initialize=x_data.columns.values
        )  # x column indices passed into set
        model.P = Set(
            initialize=y_data.columns.values
        )  # y column index passed into set

        model.x = Param(model.M, model.N, initialize=x_data.stack().to_dict())
        model.y_real = Param(model.M, model.P, initialize=y_data.stack().to_dict())

        # Define variables
        model.theta = Var(model.N, initialize=0, domain=Reals)
        model.y_predictions = Var(
            model.M, model.P, initialize=y_data.stack().to_dict(), domain=Reals
        )

        # constraint y_p = theta.X
        def xy_product(model, i, k):
            return model.y_predictions[i, k] == sum(
                model.theta[j] * model.x[i, j] for j in model.N for k in model.P
            )

        model.x_theta_product = Constraint(
            model.M, model.P, rule=xy_product, doc="Predicted value calc: y = hx"
        )

        # Cost function - RMSE
        def model_rms_error(model):
            cost_value = (1 / len(model.M)) * sum(
                ((model.y_real[i, k] - model.y_predictions[i, k]) ** 2)
                for i in model.M
                for k in model.P
            )
            return cost_value

        model.prediction_error = Objective(
            rule=model_rms_error, sense=minimize, doc="Minimum RMSE error"
        )

        instance = model
        opt = SolverFactory("ipopt")
        # TODO: This is too many iterations
        opt.options["max_iter"] = 1000
        opt.options["acceptable_tol"] = 1e-30
        # TODO: Should this be checking for a feasible solution?
        opt.solve(instance)

        # Convert theta variable into numpy array
        phi = np.zeros((len(instance.theta), 1))
        iterator = 0
        for s in instance.N:
            phi[iterator, 0] = instance.theta[s].value
            iterator += 1
        return phi

    @staticmethod
    def error_calculation(phi, x, y_data):
        """
        This function calculates the SSE and RMSE errors between the actual and predicted output values,
             ss_error = sum of squared errors / number of samples
             rmse_error = sqrt(sum of squared errors / number of samples)

        Args:
            phi             : weight vector obtained by optimization
            x               : vector of input features
            y               : actual output values

        Returns:
            ss_error        : The average sum of squared errors
            rmse_error      : The root-mean-squared error (RMSE)
            y_prediction    : Predicted values of y, y_prediction = phi.x

        """
        y_prediction = np.matmul(x, phi)
        ss_error = (1 / y_data.shape[0]) * (np.sum((y_data - y_prediction) ** 2))
        rmse_error = np.sqrt(ss_error)
        return ss_error, rmse_error, y_prediction

    @staticmethod
    def r2_calculation(y_true, y_predicted):
        """
        ``r2_calculation`` returns the :math:`R^{2}` as a measure-of-fit between the true and predicted values of the output variable.

        Args:
            y_true(NumPy Array)             : Vector of actual values of the output variable
            y_predicted(NumPy Array)        : Vector of predictions for the output variable based on the surrogate

        Returns:
            float                           : :math:`R^{2}` measure-of-fit between actual and predicted data

        """
        y_true = y_true.reshape(y_true.shape[0], 1)
        y_predicted = y_predicted.reshape(y_predicted.shape[0], 1)
        input_y_mean = np.mean(y_true, axis=0)
        ss_total = np.sum((y_true - input_y_mean) ** 2)
        ss_residual = np.sum((y_predicted - y_true) ** 2)
        r_square = 1 - (ss_residual / ss_total)
        return r_square

    def loo_error_estimation_with_rippa_method(self, sigma, lambda_reg):
        """
        The function loo_error_estimation_with_rippa_method implements the leave-one-out cross-validation (LOOCV) error for square systems

        The LOOCV error is calculated analytically using Rippa's equation:
            Error[k] = alpha[k] / inv(A[kk]),

        where:
            Error[k] = error on leaving out a particular sample k
            alpha[k] = kth radial weight based on data (kth coefficient of full data interpolation)
            A[kk] = kth diagonal element of data matrix.

        Args:
            self                          : contains, among other things, the input data
            sigma(float)                  : shape parameter for the parametric bases (Gaussian, Multiquadric, Inverse multiquadric)
            lambda_reg(float)             : regularization parameter

        Returns:
            condition_number_pure           : condition number of transformed matrix generated from the input data before regularization
            condition_number_regularized    : condition number of transformed matrix generated from the input data after regularization
            loo_error_estimate              : norm of the leave-one-out cross-validation error matrix

        For more information, see
        (1) Rippa, S. (1999) Advances in Computational Mathematics
        https://doi.org/10.1023/A:1018975909870

        (2) Mongillo M.A. (2011) Choosing Basis Functions and Shape Parameters for Radial Basis Function Methods
        https://doi.org/10.1137/11S010840

        """
        x_transformed = self.basis_generation(sigma)
        condition_number_pure = np.linalg.cond(x_transformed)

        x_regularized = x_transformed + (
            lambda_reg * np.eye(x_transformed.shape[0], x_transformed.shape[1])
        )
        condition_number_regularized = np.linalg.cond(x_regularized)

        y_train = self.y_data.reshape(self.y_data.shape[0], 1)

        # SOLVE RADIAL WEIGHTS FOR FULL X DATA
        if self.solution_method == "algebraic":
            radial_weights = self.explicit_linear_algebra_solution(
                x_regularized, y_train
            )
        elif self.solution_method == "pyomo":
            radial_weights = self.pyomo_optimization(x_regularized, y_train)
        elif self.solution_method == "bfgs":
            radial_weights = self.bfgs_parameter_optimization(x_regularized, y_train)
        radial_weights = radial_weights.reshape(radial_weights.shape[0], 1)

        # Evaluate loo-estimate with Rippa formula
        inverse_matrix = np.diag(np.linalg.pinv(x_regularized))
        error_vector = radial_weights.reshape(radial_weights.shape[0], 1) / (
            inverse_matrix.reshape(inverse_matrix.shape[0], 1)
        )
        loo_error_estimate = np.linalg.norm(error_vector)
        return condition_number_pure, condition_number_regularized, loo_error_estimate

    def leave_one_out_crossvalidation(self):
        """
        The function leave_one_out_crossvalidation determines the best hyperparameters (shape and regularization parameters) for a given RBF fitting problem.
        The function cycles through a set of predefined sets to determine the shape parameter and regularization parameter combination which yields the lowest LOOCV error.
        The LOOCV error for each (shape_parameter, regulkarization parameter) pair is evaluated by calling the function loo_error_estimation_with_rippa_method
        The pre-defined shape parameter set considers 24 irregularly spaced values ranging between 0.001 - 1000, while the regularization parameter set considers 21 values ranging between 0.00001 - 1.

        Args:
            self:                           : contains, among other things, the input data

        Returns:
            r_best(float)                 : best found shape parameter
            lambda_best(float)            : best found regularization parameter
            error_best                      : LOOCV error corresponding to be best found hynperparameters

        Note: The optimal shape parameter r_best is only evaluated for parametriuc bases (such as the Gaussian basis). For fixed basis (e.g. linear), the value is returned as zero.

        """
        # Define sigma and lambda ranges
        if (
            (self.basis_function == "gaussian")
            or (self.basis_function == "mq")
            or (self.basis_function.lower() == "imq")
        ):
            r_set = [
                0.001,
                0.002,
                0.005,
                0.0075,
                0.01,
                0.02,
                0.05,
                0.075,
                0.1,
                0.2,
                0.5,
                0.75,
                1.0,
                2.0,
                5.0,
                7.5,
                10.0,
                20.0,
                50.0,
                75.0,
                100.0,
                200.0,
                500.0,
                1000.0,
            ]
        else:
            r_set = [0]

        if self.regularization is True:
            # reg_parameter = [0.000001, 0.000005, 0.00001, 0.00005, 0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1]
            reg_parameter = [
                0.00001,
                0.00002,
                0.00005,
                0.000075,
                0.0001,
                0.0002,
                0.0005,
                0.00075,
                0.001,
                0.002,
                0.005,
                0.0075,
                0.01,
                0.02,
                0.05,
                0.075,
                0.1,
                0.2,
                0.5,
                0.75,
                1,
            ]
        elif self.regularization is False:
            reg_parameter = [0]

        machine_precision = np.finfo(float).eps

        error_vector = np.zeros((len(r_set) * len(reg_parameter), 3))
        counter = 0
        print(
            "==========================================================================================================="
        )
        for i in range(0, len(r_set)):
            sigma = r_set[i]
            for j in range(0, len(reg_parameter)):
                lambda_reg = reg_parameter[j]
                (
                    cond_no_pure,
                    cond_no_reg,
                    cv_error,
                ) = self.loo_error_estimation_with_rippa_method(sigma, lambda_reg)
                error_vector[counter, :] = [sigma, lambda_reg, cv_error]
                counter += 1
                print(
                    sigma,
                    "   |    ",
                    lambda_reg,
                    "   |    ",
                    cv_error,
                    "   |    ",
                    cond_no_pure,
                    "   |    ",
                    cond_no_pure * machine_precision,
                    "   |    ",
                    cond_no_reg,
                    "   |    ",
                    cond_no_reg * machine_precision,
                )
        minimum_value_column = np.argmin(error_vector[:, 2], axis=0)
        r_best = error_vector[minimum_value_column, 0]
        lambda_best = error_vector[minimum_value_column, 1]
        error_best = error_vector[minimum_value_column, 2]
        return r_best, lambda_best, error_best

    def training(self):
        r"""
        Main function for RBF training.

        To train the RBF:
            (1) The best values of the hyperparameters (:math:`\sigma, \lambda`) are selected via LOOCV.
            (2) The necessary basis transformation at the optimal hyperparameters is generated.
            (3) The condition number for the transformed matrix is calculated.
            (4) The optimal radial weights are evaluated using the selected optimization method.
            (5) The training predictions, prediction errors and r-square coefficient of fit are evaluated by calling the methods ``error_calculation`` and ``r2_calculation``
            (6) A results object is generated by calling the ResultsReport class.

        The LOOCV error for each (:math:`\sigma, \lambda`) pair is evaluated by calling the function ``loo_error_estimation_with_rippa_method``.

        The pre-defined shape parameter set considers 24 irregularly spaced values ranging between 0.001 - 1000, while the regularization parameter set considers 21 values ranging between 0.00001 - 1.

        Returns:
            tuple   : self object (**results**) containing the all information about the best RBF fitting obtained, including:
                - the optimal radial weights  (**results.radial_weights**),
                - when relevant, the optimal shape parameter found :math:`\sigma` (**results.sigma**),
                - when relevant, the optimal regularization parameter found :math:`\lambda` (**results.regularization**),
                - the RBF predictions for the training data (**results.output_predictions**), and
                - the :math:`R^{2}` value on the training data (**results.R2**)

        """

        # Determine best r value
        best_r_value, best_lambda_param, _ = self.leave_one_out_crossvalidation()

        # Generate x matrix
        x_transformed = self.basis_generation(best_r_value)
        x_transformed = x_transformed + (
            best_lambda_param * np.eye(x_transformed.shape[0], x_transformed.shape[1])
        )
        x_condition_number = np.linalg.cond(x_transformed)

        if self.solution_method == "algebraic":
            radial_weights = self.explicit_linear_algebra_solution(
                x_transformed, self.y_data
            )
        elif self.solution_method == "pyomo":
            radial_weights = self.pyomo_optimization(x_transformed, self.y_data)
        elif self.solution_method == "bfgs":
            radial_weights = self.bfgs_parameter_optimization(
                x_transformed, self.y_data
            )
        radial_weights = radial_weights.reshape(radial_weights.shape[0], 1)

        (
            training_ss_error,  # pylint: disable=unused-variable
            rmse_error,
            y_training_predictions_scaled,
        ) = self.error_calculation(
            radial_weights, self.basis_generation(best_r_value), self.y_data
        )
        r_square = self.r2_calculation(self.y_data, y_training_predictions_scaled)
        y_training_predictions = self.data_min[
            0, -1
        ] + y_training_predictions_scaled * (
            self.data_max[0, -1] - self.data_min[0, -1]
        )

        # Results
        self.weights = radial_weights
        self.sigma = best_r_value
        self.regularization_parameter = best_lambda_param
        self.rmse = rmse_error
        self.output_predictions = y_training_predictions
        self.condition_number = x_condition_number
        self.R2 = r_square
        self.x_data_min = self.data_min[:, :-1]
        self.x_data_max = self.data_max[:, :-1]
        self.y_data_min = self.data_min[:, -1]
        self.y_data_max = self.data_max[:, -1]
        if x_condition_number < (1 / np.finfo(float).eps):
            self.solution_status = "ok"
        else:
            _log.warning(
                "The parameter matrix A in A.x=B is ill-conditioned (condition number > 1e10). The solution returned may be inaccurate or unstable - inspect rmse error. Regularization (if not already done) may improve solution"
            )
            self.solution_status = "unstable solution"

        self.pickle_save({"model": self})
        return self

    def predict_output(self, x_data):
        """

        The ``predict_output`` method generates output predictions for input data x_data based a previously generated RBF fitting.

        Args:
            x_data(NumPy Array)    : Designs for which the output is to be evaluated/predicted.

        Returns:
             Numpy Array    : Output variable predictions based on the rbf fit.

        """
        radial_weights = self.weights
        centres_matrix = self.centres
        r = self.sigma
        scale = self.x_data_max - self.x_data_min
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_data - self.x_data_min) / scale
        x_data = x_pred_scaled.reshape(x_data.shape)

        basis_vector = np.zeros((x_data.shape[0], centres_matrix.shape[0]))
        # Calculate distances from centres
        for i in range(0, centres_matrix.shape[0]):
            basis_vector[:, i] = np.sqrt(
                np.sum(((x_data - centres_matrix[i, :]) ** 2), axis=1)
            )
        # Initialization of x_transformed
        x_transformed = np.zeros((basis_vector.shape[0], basis_vector.shape[1]))

        # Transform X
        if self.basis_function == "gaussian":
            x_transformed = RadialBasisFunctions.gaussian_basis_transformation(
                basis_vector, r
            )
        elif self.basis_function == "linear":
            x_transformed = RadialBasisFunctions.linear_transformation(basis_vector)
        elif self.basis_function == "cubic":
            x_transformed = RadialBasisFunctions.cubic_transformation(basis_vector)
        elif self.basis_function == "mq":
            x_transformed = RadialBasisFunctions.multiquadric_basis_transformation(
                basis_vector, r
            )
        elif self.basis_function == "imq":
            x_transformed = (
                RadialBasisFunctions.inverse_multiquadric_basis_transformation(
                    basis_vector, r
                )
            )
        elif self.basis_function == "spline":
            x_transformed = RadialBasisFunctions.thin_plate_spline_transformation(
                basis_vector
            )

        # Add regularization shifting?
        x_transformed = x_transformed + (
            0 * np.eye(x_transformed.shape[0], x_transformed.shape[1])
        )
        # x_transformed = x_transformed + (lambda_reg * np.eye(x_transformed.shape[0], x_transformed.shape[1]))
        y_prediction_scaled = np.matmul(x_transformed, radial_weights)
        y_prediction_unscaled = self.y_data_min + y_prediction_scaled * (
            self.y_data_max - self.y_data_min
        )
        return y_prediction_unscaled

    def generate_expression(self, variable_list):
        """
        The ``generate_expression`` method returns the Pyomo expression for the RBF model trained.

        The expression is constructed based on the supplied list of variables **variable_list** and the results of the previous RBF training process.

        Args:
            variable_list(list)           : List of input variables to be used in generating expression. This can be the a list generated from the output of ``get_feature_vector``. The user can also choose to supply a new list of the appropriate length.

        Returns:
            Pyomo Expression              : Pyomo expression of the RBF model based on the variables provided in **variable_list**

        """
        t1 = np.array([variable_list], dtype="object")
        # Reshaping of variable array is necessary when input variables are Pyomo scalar variables
        t1 = t1.reshape(1, len(variable_list)) if t1.ndim > 2 else t1

        basis_vector = []
        # Calculate distances from centres
        for i in range(0, self.centres.shape[0]):
            ans = 0
            for j in range(0, self.centres.shape[1]):
                ans += (
                    (
                        (t1[0, j] - self.x_data_min[0, j])
                        / (self.x_data_max[0, j] - self.x_data_min[0, j])
                    )
                    - self.centres[i, j]
                ) ** 2
            eucl_d = ans**0.5
            basis_vector.append(eucl_d)
        rbf_terms_list = []
        if self.basis_function == "linear":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(
                    RadialBasisFunctions.linear_transformation(basis_vector[k])
                )
        elif self.basis_function == "cubic":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(
                    RadialBasisFunctions.cubic_transformation(basis_vector[k])
                )
        elif self.basis_function == "gaussian":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(exp(-1 * ((self.sigma * basis_vector[k]) ** 2)))
        elif self.basis_function == "mq":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(
                    (((basis_vector[k] * self.sigma) ** 2) + 1) ** 0.5
                )
        elif self.basis_function == "imq":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(
                    1 / ((((basis_vector[k] * self.sigma) ** 2) + 1) ** 0.5)
                )
        elif self.basis_function == "spline":
            for k in range(0, len(basis_vector)):
                rbf_terms_list.append(((basis_vector[k] ** 2) * log(basis_vector[k])))

        rbf_terms_array = np.asarray(rbf_terms_list)
        rbf_expr = self.y_data_min[0]
        rbf_expr += (self.y_data_max[0] - self.y_data_min[0]) * sum(
            w * t
            for w, t in zip(
                np.nditer(self.weights), np.nditer(rbf_terms_array, flags=["refs_ok"])
            )
        )
        return rbf_expr

    def get_feature_vector(self):
        """

        The ``get_feature_vector`` method generates the list of regression features from the column headers of the input dataset.

        Returns:
            Pyomo IndexedParam  : An indexed parameter list of the variables supplied in the original data

        **Example:**

        .. code-block:: python

            # Create a small dataframe with three columns ('one', 'two', 'three') and two rows (A, B)
            >>> xy_data = pd.DataFrame.from_items([('A', [1, 2, 3]), ('B', [4, 5, 6])], orient='index', columns=['one', 'two', 'three'])

            # Initialize the **RadialBasisFunctions** class with a linear kernel and print the column headers for the variables
            >>> f = RadialBasisFunctions(xy_data, basis_function='linear')
            >>> p = f.get_feature_vector()
            >>> for i in p.keys():
            >>>     print(i)
            one
            two

        """
        p = Param(self.x_data_columns, mutable=True, initialize=0)
        p.index_set().construct()
        p.construct()
        self.feature_list = p
        return p

    def pickle_save(self, solutions):
        """
        The pickle_save method saves the results of the run in a pickle object.
        """
        try:
            filehandler = open(self.filename, "wb")
            pickle.dump(solutions, filehandler)
            print("\nResults saved in ", str(self.filename))
        except:
            raise IOError("File could not be saved.")

    @staticmethod
    def pickle_load(solution_file):
        """
        pickle_load loads the results of a saved run 'file.obj'.

        Input arguments:
                solution_file            : Pickle object file containing previous solution to be loaded.

        """
        try:
            filehandler = open(solution_file, "rb")
            return pickle.load(filehandler)
        except:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("File could not be loaded.")

    def parity_residual_plots(self):
        """

        inputs:

        Returns:

        """

        fig1 = plt.figure(figsize=(16, 9), tight_layout=True)
        ax = fig1.add_subplot(121)
        ax.plot(self.y_data_unscaled, self.y_data_unscaled, "-")
        ax.plot(self.y_data_unscaled, self.output_predictions, "o")
        ax.set_xlabel(r"True data", fontsize=12)
        ax.set_ylabel(r"Surrogate values", fontsize=12)
        ax.set_title(r"Parity plot", fontsize=12)

        ax2 = fig1.add_subplot(122)
        ax2.plot(
            self.y_data_unscaled,
            self.y_data_unscaled - self.output_predictions,
            "s",
            mfc="w",
            mec="m",
            ms=6,
        )
        ax2.axhline(y=0, xmin=0, xmax=1)
        ax2.set_xlabel(r"True data", fontsize=12)
        ax2.set_ylabel(r"Residuals", fontsize=12)
        ax2.set_title(r"Residual plot", fontsize=12)

        plt.show()

        return

    def _report(self):
        ## Will only work with Python > 3.5
        variable_headers = self.get_feature_vector()
        var_list = []
        for i in variable_headers:
            var_list.append(variable_headers[i])
        eqn = self.generate_expression(var_list)

        double_line = "=" * 120
        s = (
            f"\n{double_line}"
            f"\nResults of radial basis function run:\n"
            f"\nBasis function type               : {self.basis_function}\n"
            f"Shape parameter                    : {self.sigma}\n"
            f"Regularization parameter           : {self.regularization_parameter}\n"
            f"Number of terms in RBF model       : {self.weights.size + 1}\n"  # The additional term is y_min
            f"\nRBF Expression:\n"
            f"--------------------------\n"
            f"\n{eqn}\n"
            f"--------------------------\n"
            f"\nModel training errors:"
            f"\n-----------------------\n"
            f"Mean Squared Error (MSE)         : {self.rmse ** 2}\n"
            f"Root Mean Squared Error (RMSE)   : {self.rmse}\n"
            f"Goodness of fit (R2)             : {self.R2}\n"
            f"\n{double_line}"
        )
        return s

    def print_report(self):
        s = self._report()
        print(s)

    def _repr_pretty_(self, p, cycle=False):
        s = self._report()
        p.text(s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Generic template for a translator block.
"""
# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
from idaes.core.initialization import BlockTriangularizationInitializer

__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Translator")
class TranslatorData(UnitModelBlockData):
    """
    Standard Translator Block Class
    """

    # Set default initializer
    # TODO: For now, use Block Triangularization as the default
    default_initializer = BlockTriangularizationInitializer

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Translator blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Translator blocks do not contain holdup.""",
        ),
    )
    CONFIG.declare(
        "outlet_state_defined",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Indicated whether outlet state will be fully defined",
            doc="""Indicates whether unit model will fully define outlet state.
If False, the outlet property package will enforce constraints such as sum
of mole fractions and phase equilibrium.
**default** - True.
**Valid values:** {
**True** - outlet state will be fully defined,
**False** - outlet property package should enforce summation and equilibrium
constraints.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Indicates whether outlet is in phase equilibrium",
            doc="""Indicates whether outlet property package should enforce
phase equilibrium constraints.
**default** - False.
**Valid values:** {
**True** - outlet property package should calculate phase equilibrium,
**False** - outlet property package should notcalculate phase equilibrium.}
""",
        ),
    )
    CONFIG.declare(
        "inlet_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for incoming stream",
            doc="""Property parameter object used to define property
calculations for the incoming stream,
**default** - None.
**Valid values:** {
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "inlet_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package "
            "of the incoming stream",
            doc="""A ConfigBlock with arguments to be passed to the property
block associated with the incoming stream,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for outgoing stream",
            doc="""Property parameter object used to define property
calculations for the outgoing stream,
**default** - None.
**Valid values:** {
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "outlet_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package "
            "of the outgoing stream",
            doc="""A ConfigBlock with arguments to be passed to the property
block associated with the outgoing stream,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TranslatorData, self).build()

        # Check construction argument consistency
        if self.config.outlet_state_defined and self.config.has_phase_equilibrium:
            raise ConfigurationError(
                "{} cannot calculate phase equilibrium (has_phase_equilibrium "
                "= True) when outlet state is set to be fully defined ("
                "outlet_state_defined = True).".format(self.name)
            )

        # Add State Blocks
        self.properties_in = self.config.inlet_property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in incoming stream",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.inlet_property_package_args
        )

        self.properties_out = self.config.outlet_property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in outgoing stream",
            defined_state=self.config.outlet_state_defined,
            has_phase_equilibrium=self.config.has_phase_equilibrium,
            **self.config.outlet_property_package_args
        )

        # Add outlet port
        self.add_port(name="inlet", block=self.properties_in, doc="Inlet Port")
        self.add_port(name="outlet", block=self.properties_out, doc="Outlet Port")

    def initialize_build(
        blk,
        state_args_in=None,
        state_args_out=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        This method calls the initialization method of the state blocks.

        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                            property package (to provide an initial state for
                            initialization (see documentation of the specific
                            property package) (default = None).
            state_args_out : a dict of arguments to be passed to the outlet
                             property package (to provide an initial state for
                             initialization (see documentation of the specific
                             property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state block
        flags = blk.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_in,
            hold_state=True,
        )

        blk.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        if degrees_of_freedom(blk) == 0:
            with idaeslog.solver_log(init_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)

            init_log.info("Initialization Complete {}.".format(idaeslog.condition(res)))
        else:
            init_log.warning(
                "Initialization incomplete. Degrees of freedom "
                "were not zero. Please provide sufficient number "
                "of constraints linking the state variables "
                "between the two state blocks."
            )

        blk.properties_in.release_state(flags=flags, outlvl=outlvl)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Created on Thu Mar 18 15:05:41 2021

@author: andrew
"""
import pytest

import pyomo.environ as pe
import idaes.core
import idaes.models.unit_models
import idaes.models.properties.swco2
from pyomo.network import Arc
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.fixture(scope="function")
def model():
    m = pe.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock()
    m.fs.properties = idaes.models.properties.swco2.SWCO2ParameterBlock()
    m.fs.heater = idaes.models.unit_models.Heater(
        dynamic=False, property_package=m.fs.properties, has_pressure_change=True
    )
    m.fs.heater2 = idaes.models.unit_models.Heater(
        dynamic=False, property_package=m.fs.properties, has_pressure_change=True
    )
    m.fs.stream = Arc(source=m.fs.heater.outlet, destination=m.fs.heater2.inlet)

    return m


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_expand_arcs_and_clone(model):
    # Check that port references were attached to model
    assert hasattr(model.fs.heater, "_enth_mol_inlet_ref")
    assert hasattr(model.fs.heater, "_flow_mol_inlet_ref")
    assert hasattr(model.fs.heater, "_pressure_inlet_ref")
    assert hasattr(model.fs.heater, "_enth_mol_outlet_ref")
    assert hasattr(model.fs.heater, "_flow_mol_outlet_ref")
    assert hasattr(model.fs.heater, "_pressure_outlet_ref")

    assert hasattr(model.fs.heater2, "_enth_mol_inlet_ref")
    assert hasattr(model.fs.heater2, "_flow_mol_inlet_ref")
    assert hasattr(model.fs.heater2, "_pressure_inlet_ref")
    assert hasattr(model.fs.heater2, "_enth_mol_outlet_ref")
    assert hasattr(model.fs.heater2, "_flow_mol_outlet_ref")
    assert hasattr(model.fs.heater2, "_pressure_outlet_ref")

    assert model.fs.stream._expanded_block is None

    m2 = pe.TransformationFactory("network.expand_arcs").create_using(model)

    # Check that Arcs were expanded
    assert isinstance(m2.fs.stream._expanded_block, pe.Block)
    assert model.fs.stream._expanded_block is None


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_clone(model):
    m2 = model.clone()
    assert (
        m2.fs.heater.inlet.flow_mol[0]
        is m2.fs.heater.control_volume.properties_in[0].flow_mol
    )

    assert not (
        m2.fs.heater.inlet.flow_mol[0]
        is model.fs.heater.control_volume.properties_in[0].flow_mol
    )

    pe.TransformationFactory("network.expand_arcs").apply_to(m2)

    # Check that Arcs were expanded
    assert isinstance(m2.fs.stream._expanded_block, pe.Block)
    assert model.fs.stream._expanded_block is None



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
NRTL property model for a benzene-toluene mixture.
The example model is from the IDAES tutorial,
https://github.com/IDAES/examples-pse/blob/main/src/Tutorials/Advanced/ParamEst/
"""
from idaes.core import FlowsheetBlock
from idaes.models.unit_models import Flash
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
import idaes.logger as idaeslog
from pyomo.environ import *


def NRTL_model(data):
    """This function generates an instance of the NRTL Pyomo model using 'data' as the input argument

    Parameters
    ----------
    data: pandas DataFrame, list of dictionaries, or list of json file names
        Data that is used to build an instance of the Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for estimating parameters and covariance
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="NRTL"
    )
    m.fs.flash = Flash(property_package=m.fs.properties)

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Initialize the flash unit
    m.fs.flash.initialize(outlvl=idaeslog.INFO_LOW)

    # Fix at actual temperature
    m.fs.flash.inlet.temperature.fix(float(data["temperature"]))

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)

    # Return initialized flash model
    return m


def NRTL_model_opt():
    """This function generates an instance of the NRTL Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="NRTL"
    )
    m.fs.flash = Flash(property_package=m.fs.properties)

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Initialize the flash unit
    m.fs.flash.initialize(outlvl=idaeslog.INFO_LOW)

    # Fix at actual temperature
    m.fs.flash.inlet.temperature.fix(float(368))

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)

    # To use kaug
    # objective function required
    # need to unfix the variables
    m.obj = Objective(
        expr=0 * m.fs.properties.tau["benzene", "toluene"]
        + exp(
            -m.fs.properties.alpha["toluene", "benzene"].value
            * m.fs.properties.tau["toluene", "benzene"]
        ),
        sense=minimize,
    )
    m.fs.properties.tau["benzene", "toluene"].fixed = False  # To use kaug
    m.fs.properties.tau["toluene", "benzene"].fixed = False
    return m


def NRTL_model_opt_infeasible():
    """This function generates an instance of the NRTL Pyomo model

    Returns
    -------
    m: an instance of the Pyomo model
        for uncertainty propagation
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="NRTL"
    )
    m.fs.flash = Flash(property_package=m.fs.properties)

    # Initialize at a certain inlet condition
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

    # Set Flash unit specifications
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)

    # Fix NRTL specific variables
    # alpha values (set at 0.3)
    m.fs.properties.alpha["benzene", "benzene"].fix(0)
    m.fs.properties.alpha["benzene", "toluene"].fix(0.3)
    m.fs.properties.alpha["toluene", "toluene"].fix(0)
    m.fs.properties.alpha["toluene", "benzene"].fix(0.3)

    # initial tau values
    m.fs.properties.tau["benzene", "benzene"].fix(0)
    m.fs.properties.tau["benzene", "toluene"].fix(0.1690)
    m.fs.properties.tau["toluene", "toluene"].fix(0)
    m.fs.properties.tau["toluene", "benzene"].fix(-0.1559)

    # Set bounds on variables to be estimated
    m.fs.properties.tau["benzene", "toluene"].setlb(-5)
    m.fs.properties.tau["benzene", "toluene"].setub(5)

    m.fs.properties.tau["toluene", "benzene"].setlb(-5)
    m.fs.properties.tau["toluene", "benzene"].setub(5)

    # To use kaug
    # objective function required
    # need to unfix the variables
    m.obj = Objective(
        expr=0 * m.fs.properties.tau["benzene", "toluene"]
        + exp(
            -m.fs.properties.alpha["toluene", "benzene"].value
            * m.fs.properties.tau["toluene", "benzene"]
        ),
        sense=minimize,
    )
    m.fs.properties.tau["benzene", "toluene"].fixed = False
    m.fs.properties.tau["toluene", "benzene"].fixed = False
    return m



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Ideal + Wilson Liquid activity coefficient state block;
only tests for construction.

Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import ConcreteModel
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.core.util.model_statistics import degrees_of_freedom


# -----------------------------------------------------------------------------
@pytest.fixture(scope="module")
def m():
    # Create a flowsheet for test
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # vapor-liquid (Wilson)
    m.fs.properties_Wilson_vl = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_vl = m.fs.properties_Wilson_vl.build_state_block(
        defined_state=True
    )

    # liquid only (Wilson)
    m.fs.properties_Wilson_l = BTXParameterBlock(
        valid_phase="Liq", activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_l = m.fs.properties_Wilson_l.build_state_block(
        has_phase_equilibrium=False, defined_state=True
    )

    # vapour only (Wilson)
    m.fs.properties_Wilson_v = BTXParameterBlock(
        valid_phase="Vap", activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_v = m.fs.properties_Wilson_v.build_state_block(
        has_phase_equilibrium=False, defined_state=True
    )

    return m


@pytest.mark.unit
def test_build_inlet_state_block(m):
    assert len(m.fs.properties_Wilson_vl.config) == 4

    # vapor-liquid (Wilson)
    assert m.fs.properties_Wilson_vl.config.valid_phase == (
        "Vap",
        "Liq",
    ) or m.fs.properties_Wilson_vl.config.valid_phase == ("Liq", "Vap")
    assert len(m.fs.properties_Wilson_vl.phase_list) == 2
    assert m.fs.properties_Wilson_vl.phase_list == ["Liq", "Vap"]
    assert m.fs.state_block_Wilson_vl.config.defined_state
    assert hasattr(m.fs.state_block_Wilson_vl, "eq_phase_equilibrium")
    assert hasattr(m.fs.state_block_Wilson_vl, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_Wilson_vl, "eq_mol_frac_out")

    # liquid only (Wilson)
    assert len(m.fs.properties_Wilson_l.config) == 4

    assert m.fs.properties_Wilson_l.config.valid_phase == "Liq"
    assert len(m.fs.properties_Wilson_l.phase_list) == 1
    assert m.fs.properties_Wilson_l.phase_list == ["Liq"]
    assert m.fs.state_block_Wilson_l.config.defined_state
    assert not hasattr(m.fs.state_block_Wilson_l, "eq_phase_equilibrium")
    assert not hasattr(m.fs.state_block_Wilson_l, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_Wilson_l, "eq_mol_frac_out")

    # vapor only (Wilson)
    assert len(m.fs.properties_Wilson_v.config) == 4

    assert m.fs.properties_Wilson_v.config.valid_phase == "Vap"
    assert len(m.fs.properties_Wilson_v.phase_list) == 1
    assert m.fs.properties_Wilson_v.phase_list == ["Vap"]
    assert m.fs.state_block_Wilson_v.config.defined_state
    assert not hasattr(m.fs.state_block_Wilson_v, "eq_phase_equilibrium")
    assert not hasattr(m.fs.state_block_Wilson_v, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_Wilson_v, "eq_mol_frac_out")


@pytest.mark.unit
def test_setInputs_inlet_state_block(m):
    # vapor-liquid (Wilson)
    m.fs.state_block_Wilson_vl.flow_mol.fix(1)
    m.fs.state_block_Wilson_vl.temperature.fix(368)
    m.fs.state_block_Wilson_vl.pressure.fix(101325)
    m.fs.state_block_Wilson_vl.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_Wilson_vl.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_Wilson_vl) == 4

    # Fix Wilson specific variables
    m.fs.properties_Wilson_vl.vol_mol_comp.fix()
    m.fs.properties_Wilson_vl.tau.fix()

    assert degrees_of_freedom(m.fs.state_block_Wilson_vl) == 0

    # liquid only (Wilson)
    m.fs.state_block_Wilson_l.flow_mol.fix(1)
    m.fs.state_block_Wilson_l.temperature.fix(368)
    m.fs.state_block_Wilson_l.pressure.fix(101325)
    m.fs.state_block_Wilson_l.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_Wilson_l.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_Wilson_l) == 0

    # vapour only (Wilson)
    m.fs.state_block_Wilson_v.flow_mol.fix(1)
    m.fs.state_block_Wilson_v.temperature.fix(368)
    m.fs.state_block_Wilson_v.pressure.fix(101325)
    m.fs.state_block_Wilson_v.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_Wilson_v.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_Wilson_v) == 0


@pytest.fixture(scope="module")
def m2():
    # Create a flowsheet for test
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # vapor-liquid (Wilson)
    m.fs.properties_Wilson_vl = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_vl = m.fs.properties_Wilson_vl.build_state_block(
        defined_state=True
    )

    # liquid only (Wilson)
    m.fs.properties_Wilson_l = BTXParameterBlock(
        valid_phase="Liq", activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_l = m.fs.properties_Wilson_l.build_state_block(
        has_phase_equilibrium=False, defined_state=True
    )

    # vapour only (Wilson)
    m.fs.properties_Wilson_v = BTXParameterBlock(
        valid_phase="Vap", activity_coeff_model="Wilson"
    )
    m.fs.state_block_Wilson_v = m.fs.properties_Wilson_v.build_state_block(
        has_phase_equilibrium=False, defined_state=True
    )

    # Create a flowsheet object to test outlet state blocks
    m.fs1 = FlowsheetBlock(dynamic=False)

    # vapor-liquid (Wilson)
    m.fs1.properties_Wilson_vl = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Wilson"
    )
    m.fs1.state_block_Wilson_vl = m.fs1.properties_Wilson_vl.build_state_block(
        defined_state=False
    )

    # liquid only (Wilson)
    m.fs1.properties_Wilson_l = BTXParameterBlock(
        valid_phase="Liq", activity_coeff_model="Wilson"
    )
    m.fs1.state_block_Wilson_l = m.fs1.properties_Wilson_l.build_state_block(
        has_phase_equilibrium=False, defined_state=False
    )

    # vapour only (Wilson)
    m.fs1.properties_Wilson_v = BTXParameterBlock(
        valid_phase="Vap", activity_coeff_model="Wilson"
    )
    m.fs1.state_block_Wilson_v = m.fs1.properties_Wilson_v.build_state_block(
        has_phase_equilibrium=False, defined_state=False
    )

    return m


@pytest.mark.unit
def test_build_outlet_state_block(m2):
    m = m2
    assert len(m.fs.properties_Wilson_vl.config) == 4

    # vapor-liquid (Wilson)
    assert m.fs1.properties_Wilson_vl.config.valid_phase == (
        "Vap",
        "Liq",
    ) or m.fs1.properties_Wilson_vl.config.valid_phase == ("Liq", "Vap")
    assert len(m.fs1.properties_Wilson_vl.phase_list) == 2
    assert m.fs1.properties_Wilson_vl.phase_list == ["Liq", "Vap"]
    assert not m.fs1.state_block_Wilson_vl.config.defined_state
    assert hasattr(m.fs1.state_block_Wilson_vl, "eq_phase_equilibrium")
    assert hasattr(m.fs1.state_block_Wilson_vl, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_Wilson_vl, "eq_mol_frac_out")

    # liquid only (Wilson)
    assert len(m.fs1.properties_Wilson_l.config) == 4

    assert m.fs1.properties_Wilson_l.config.valid_phase == "Liq"
    assert len(m.fs1.properties_Wilson_l.phase_list) == 1
    assert m.fs1.properties_Wilson_l.phase_list == ["Liq"]
    assert not m.fs1.state_block_Wilson_l.config.defined_state
    assert not hasattr(m.fs1.state_block_Wilson_l, "eq_phase_equilibrium")
    assert not hasattr(m.fs1.state_block_Wilson_l, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_Wilson_l, "eq_mol_frac_out")

    # vapour only (Wilson)
    assert len(m.fs1.properties_Wilson_v.config) == 4

    assert m.fs1.properties_Wilson_v.config.valid_phase == "Vap"
    assert len(m.fs1.properties_Wilson_v.phase_list) == 1
    assert m.fs1.properties_Wilson_v.phase_list == ["Vap"]
    assert not m.fs1.state_block_Wilson_v.config.defined_state
    assert not hasattr(m.fs1.state_block_Wilson_v, "eq_phase_equilibrium")
    assert not hasattr(m.fs1.state_block_Wilson_v, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_Wilson_v, "eq_mol_frac_out")


@pytest.mark.unit
def test_setInputs_outlet_state_block(m2):
    m = m2

    # vapor-liquid (Wilson)
    m.fs1.state_block_Wilson_vl.flow_mol.fix(1)
    m.fs1.state_block_Wilson_vl.temperature.fix(368)
    m.fs1.state_block_Wilson_vl.pressure.fix(101325)
    m.fs1.state_block_Wilson_vl.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_Wilson_vl) == 4

    # liquid only (Wilson)
    m.fs1.state_block_Wilson_l.flow_mol.fix(1)
    m.fs1.state_block_Wilson_l.temperature.fix(368)
    m.fs1.state_block_Wilson_l.pressure.fix(101325)
    m.fs1.state_block_Wilson_l.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_Wilson_l) == 0

    # vapour only (Wilson)
    m.fs1.state_block_Wilson_v.flow_mol.fix(1)
    m.fs1.state_block_Wilson_v.temperature.fix(368)
    m.fs1.state_block_Wilson_v.pressure.fix(101325)
    m.fs1.state_block_Wilson_v.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_Wilson_v) == 0


@pytest.mark.integration
def test_units_consistent(m2):
    m = m2
    assert_units_consistent(m.fs)
    assert_units_consistent(m.fs1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for setting up FTPx as the state variables in a generic property
package
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from types import MethodType
from pyomo.environ import (
    Constraint,
    Expression,
    NonNegativeReals,
    Var,
    value,
    units as pyunits,
)

from idaes.core import MaterialFlowBasis, MaterialBalanceType, EnergyBalanceType
from idaes.models.properties.modular_properties.base.utility import (
    get_bounds_from_config,
    get_method,
    GenericPropertyPackageError,
)
from idaes.models.properties.modular_properties.base.utility import (
    identify_VL_component_list,
)
from idaes.models.properties.modular_properties.phase_equil.henry import (
    HenryType,
    henry_equilibrium_ratio,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from .electrolyte_states import define_electrolyte_state, calculate_electrolyte_scaling


# Set up logger
_log = idaeslog.getLogger(__name__)


def set_metadata(b):
    # This is the default assumption for state vars, so we don't need to change
    # the metadata dict
    pass


def define_state(b):
    # FTPx formulation always requires a flash, so set flag to True
    # TODO: should have some checking to make sure developers implement this properly
    b.always_flash = True

    units = b.params.get_metadata().derived_units

    # Check that only necessary state_bounds are defined
    expected_keys = ["flow_mol", "temperature", "pressure"]
    if (
        b.params.config.state_bounds is not None
        and any(b.params.config.state_bounds.keys()) not in expected_keys
    ):
        for k in b.params.config.state_bounds.keys():
            if "mole_frac" in k:
                _log.warning(
                    "{} - found state_bounds argument for {}."
                    " Mole fraction bounds are set automatically and "
                    "this argument will be ignored.".format(b.name, k)
                )
            elif k not in expected_keys:
                raise ConfigurationError(
                    "{} - found unexpected state_bounds key {}. Please ensure "
                    "bounds are provided only for expected state variables "
                    "and that you have typed the variable names correctly.".format(
                        b.name, k
                    )
                )

    # Get bounds and initial values from config args
    f_bounds, f_init = get_bounds_from_config(b, "flow_mol", units.FLOW_MOLE)
    t_bounds, t_init = get_bounds_from_config(b, "temperature", units.TEMPERATURE)
    p_bounds, p_init = get_bounds_from_config(b, "pressure", units.PRESSURE)

    # Add state variables
    b.flow_mol = Var(
        initialize=f_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc=" Total molar flowrate",
        units=units.FLOW_MOLE,
    )
    b.mole_frac_comp = Var(
        b.component_list,
        bounds=(1e-20, 1.001),
        initialize=1 / len(b.component_list),
        doc="Mixture mole fractions",
        units=pyunits.dimensionless,
    )
    b.pressure = Var(
        initialize=p_init,
        domain=NonNegativeReals,
        bounds=p_bounds,
        doc="State pressure",
        units=units.PRESSURE,
    )
    b.temperature = Var(
        initialize=t_init,
        domain=NonNegativeReals,
        bounds=t_bounds,
        doc="State temperature",
        units=units.TEMPERATURE,
    )

    # Add supporting variables
    if f_init is None:
        fp_init = None
    else:
        fp_init = f_init / len(b.phase_list)

    b.flow_mol_phase = Var(
        b.phase_list,
        initialize=fp_init,
        domain=NonNegativeReals,
        bounds=f_bounds,
        doc="Phase molar flow rates",
        units=units.FLOW_MOLE,
    )

    b.mole_frac_phase_comp = Var(
        b.phase_component_set,
        initialize=1 / len(b.component_list),
        bounds=(1e-20, 1.001),
        doc="Phase mole fractions",
        units=pyunits.dimensionless,
    )

    def Fpc_expr(b, p, j):
        return b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, j]

    b.flow_mol_phase_comp = Expression(
        b.phase_component_set, rule=Fpc_expr, doc="Phase-component molar flowrates"
    )

    b.phase_frac = Var(
        b.phase_list,
        initialize=1.0 / len(b.phase_list),
        bounds=(0, None),
        doc="Phase fractions",
        units=pyunits.dimensionless,
    )

    # Add electrolyte state vars if required
    if b.params._electrolyte:
        define_electrolyte_state(b)

    # Add supporting constraints
    if b.config.defined_state is False:
        # applied at outlet only
        b.sum_mole_frac_out = Constraint(
            expr=1.0 == sum(b.mole_frac_comp[i] for i in b.component_list)
        )

    if len(b.phase_list) == 1:

        def rule_total_mass_balance(b):
            return b.flow_mol_phase[b.phase_list.first()] == b.flow_mol

        b.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return (
                b.mole_frac_comp[i] == b.mole_frac_phase_comp[b.phase_list.first(), i]
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_phase_frac(b, p):
            return b.phase_frac[p] == 1.0

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    elif len(b.phase_list) == 2:
        # For two phase, use Rachford-Rice formulation
        def rule_total_mass_balance(b):
            return sum(b.flow_mol_phase[p] for p in b.phase_list) == b.flow_mol

        b.total_flow_balance = Constraint(rule=rule_total_mass_balance)

        def rule_comp_mass_balance(b, i):
            return b.flow_mol * b.mole_frac_comp[i] == sum(
                b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, i]
                for p in b.phase_list
                if (p, i) in b.phase_component_set
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_mole_frac(b):
            return (
                sum(
                    b.mole_frac_phase_comp[b.phase_list.first(), i]
                    for i in b.component_list
                    if (b.phase_list.first(), i) in b.phase_component_set
                )
                - sum(
                    b.mole_frac_phase_comp[b.phase_list.last(), i]
                    for i in b.component_list
                    if (b.phase_list.last(), i) in b.phase_component_set
                )
                == 0
            )

        b.sum_mole_frac = Constraint(rule=rule_mole_frac)

        def rule_phase_frac(b, p):
            return b.phase_frac[p] * b.flow_mol == b.flow_mol_phase[p]

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    else:
        # Otherwise use a general formulation
        def rule_comp_mass_balance(b, i):
            return b.flow_mol * b.mole_frac_comp[i] == sum(
                b.flow_mol_phase[p] * b.mole_frac_phase_comp[p, i]
                for p in b.phase_list
                if (p, i) in b.phase_component_set
            )

        b.component_flow_balances = Constraint(
            b.component_list, rule=rule_comp_mass_balance
        )

        def rule_mole_frac(b, p):
            return (
                sum(
                    b.mole_frac_phase_comp[p, i]
                    for i in b.component_list
                    if (p, i) in b.phase_component_set
                )
                == 1
            )

        b.sum_mole_frac = Constraint(b.phase_list, rule=rule_mole_frac)

        def rule_phase_frac(b, p):
            return b.phase_frac[p] * b.flow_mol == b.flow_mol_phase[p]

        b.phase_fraction_constraint = Constraint(b.phase_list, rule=rule_phase_frac)

    # -------------------------------------------------------------------------
    # General Methods
    def get_material_flow_terms_FTPx(b, p, j):
        """Create material flow terms for control volume."""
        return b.flow_mol_phase_comp[p, j]

    b.get_material_flow_terms = MethodType(get_material_flow_terms_FTPx, b)

    def get_enthalpy_flow_terms_FTPx(b, p):
        """Create enthalpy flow terms."""
        # enth_mol_phase probably does not exist when this is created
        # Use try/except to build flow term if not present
        try:
            eflow = b._enthalpy_flow_term
        except AttributeError:

            def rule_eflow(b, p):
                return b.flow_mol_phase[p] * b.enth_mol_phase[p]

            eflow = b._enthalpy_flow_term = Expression(b.phase_list, rule=rule_eflow)
        return eflow[p]

    b.get_enthalpy_flow_terms = MethodType(get_enthalpy_flow_terms_FTPx, b)

    def get_material_density_terms_FTPx(b, p, j):
        """Create material density terms."""
        # dens_mol_phase probably does not exist when this is created
        # Use try/except to build term if not present
        try:
            mdens = b._material_density_term
        except AttributeError:

            def rule_mdens(b, p, j):
                return b.dens_mol_phase[p] * b.mole_frac_phase_comp[p, j]

            mdens = b._material_density_term = Expression(
                b.phase_component_set, rule=rule_mdens
            )
        return mdens[p, j]

    b.get_material_density_terms = MethodType(get_material_density_terms_FTPx, b)

    def get_energy_density_terms_FTPx(b, p):
        """Create energy density terms."""
        # Density and energy terms probably do not exist when this is created
        # Use try/except to build term if not present
        try:
            edens = b._energy_density_term
        except AttributeError:

            def rule_edens(b, p):
                return b.dens_mol_phase[p] * b.energy_internal_mol_phase[p]

            edens = b._energy_density_term = Expression(b.phase_list, rule=rule_edens)
        return edens[p]

    b.get_energy_density_terms = MethodType(get_energy_density_terms_FTPx, b)

    def default_material_balance_type_FTPx():
        return MaterialBalanceType.componentTotal

    b.default_material_balance_type = default_material_balance_type_FTPx

    def default_energy_balance_type_FTPx():
        return EnergyBalanceType.enthalpyTotal

    b.default_energy_balance_type = default_energy_balance_type_FTPx

    def get_material_flow_basis_FTPx():
        return MaterialFlowBasis.molar

    b.get_material_flow_basis = get_material_flow_basis_FTPx

    def define_state_vars_FTPx(b):
        """Define state vars."""
        return {
            "flow_mol": b.flow_mol,
            "mole_frac_comp": b.mole_frac_comp,
            "temperature": b.temperature,
            "pressure": b.pressure,
        }

    b.define_state_vars = MethodType(define_state_vars_FTPx, b)

    def define_display_vars_FTPx(b):
        """Define display vars."""
        return {
            "Total Molar Flowrate": b.flow_mol,
            "Total Mole Fraction": b.mole_frac_comp,
            "Temperature": b.temperature,
            "Pressure": b.pressure,
        }

    b.define_display_vars = MethodType(define_display_vars_FTPx, b)


def state_initialization(b):
    # Need to do sanity checking of mole fractions for ideal phase fraction
    # calculations
    for j in b.component_list:
        if value(b.mole_frac_comp[j]) < 0:
            raise ValueError(
                f"Component {j} has a negative "
                f"mole fraction in block {b.name}. "
                "Check your initialization."
            )

    for p in b.phase_list:
        # Start with phase mole fractions equal to total mole fractions
        for j in b.component_list:
            if (p, j) in b.phase_component_set:
                b.mole_frac_phase_comp[p, j].value = b.mole_frac_comp[j].value

        b.flow_mol_phase[p].value = value(b.phase_frac[p] * b.flow_mol)

    if not hasattr(b.params, "_pe_pairs"):
        return
    else:
        _pe_pairs = b.params._pe_pairs

    vl_comps = []
    henry_comps = []
    init_VLE = False
    for pp in _pe_pairs:
        # Look for a VLE pair with this phase - should only be 1
        if (
            b.params.get_phase(pp[0]).is_liquid_phase()
            and b.params.get_phase(pp[1]).is_vapor_phase()
        ) or (
            b.params.get_phase(pp[1]).is_liquid_phase()
            and b.params.get_phase(pp[0]).is_vapor_phase()
        ):
            init_VLE = True
            # Get bubble and dew points
            tbub = None
            tdew = None
            if hasattr(b, "eq_temperature_bubble"):
                try:
                    tbub = b.temperature_bubble[pp].value
                except KeyError:
                    pass
            if hasattr(b, "eq_temperature_dew"):
                try:
                    tdew = b.temperature_dew[pp].value
                except KeyError:
                    pass
            if len(vl_comps) > 0:
                # More than one VLE. Just use the default initialization for
                # now
                init_VLE = False
            (
                l_phase,
                v_phase,
                vl_comps,
                henry_comps,
                l_only_comps,
                v_only_comps,
            ) = identify_VL_component_list(b, pp)
            pp_VLE = pp

    if init_VLE:
        henry_mole_frac = []
        henry_conc = []
        henry_other = []
        K = {}

        for j in henry_comps:
            henry_type = b.params.get_component(j).config.henry_component[l_phase][
                "type"
            ]
            if henry_type in {HenryType.Hxp, HenryType.Kpx}:
                # Quick and dirty way of getting a Henry constant in the right
                # form/units is by temporarily setting its mole fraction to
                # one, then calling the henry_pressure method. Reset mole
                # fraction to original value to "do no harm"
                henry_mole_frac.append(j)
                K[j] = value(henry_equilibrium_ratio(b, l_phase, j))

            elif henry_type in {HenryType.Hcp, HenryType.Kpc}:
                # Can't evaluate concentration without density, can't evaluate
                # density without liquid phase concentration. Treat as
                # noncondensible and correct value during a second pass
                henry_conc.append(j)
            else:
                _log.warning(
                    "{} - Components with unsupported Henry's Law "
                    "type {} are present, and will be treated as "
                    "noncondensible by the vapor-liquid equilibrium "
                    "initialization for {} state variables.".format(
                        b.name, str(henry_type), str(b.params.config.state_definition)
                    )
                )
                henry_other.append(j)
        for j in vl_comps:
            try:
                K[j] = value(
                    get_method(b, "pressure_sat_comp", j)(
                        b, b.params.get_component(j), b.temperature
                    )
                    / b.pressure
                )
            except GenericPropertyPackageError:
                # No method for calculating Psat, use default values
                K = None
                break

    # Default is no initialization of VLE
    vap_frac = None
    if init_VLE:
        raoult_init = False
        if tdew is not None and b.temperature.value > tdew:
            # Pure vapour
            vap_frac = 1 - 1e-5
        elif tbub is not None and b.temperature.value < tbub:
            # Pure liquid
            vap_frac = 1e-5
        elif tbub is not None and tdew is not None:
            # Two-phase with bounds two-phase region
            # Thanks to Rahul Gandhi for the method
            vap_frac = value(b.temperature - tbub) / (tdew - tbub)
        elif K is not None:
            raoult_init = True
            vap_frac = _modified_rachford_rice(
                b,
                K,
                vl_comps + henry_mole_frac,
                l_only_comps,
                v_only_comps + henry_conc + henry_other,
            )
        # else: No way to estimate phase fraction, do nothing

    if vap_frac is not None:
        b.phase_frac[v_phase] = vap_frac
        b.phase_frac[l_phase] = 1 - vap_frac

        for p in pp_VLE:
            b.flow_mol_phase[p].value = value(b.phase_frac[p] * b.flow_mol)

        # If dew/bubble calculations or MRR gives a nearly single-phase vapor
        # fraction, try to initialize mole fraction using dew/bubble comps
        if tbub is not None and vap_frac < 1e-4:
            # Pure liquid
            for j in b.component_list:
                if (l_phase, j) in b.phase_component_set:
                    b.mole_frac_phase_comp[l_phase, j].value = b.mole_frac_comp[j].value
                if (v_phase, j) in b.phase_component_set:
                    b.mole_frac_phase_comp[v_phase, j].value = b._mole_frac_tbub[
                        pp_VLE, j
                    ].value
        elif tdew is not None and vap_frac > 1 - 1e-4:
            # Pure Vapor
            for j in b.component_list:
                if (v_phase, j) in b.phase_component_set:
                    b.mole_frac_phase_comp[v_phase, j].value = b.mole_frac_comp[j].value
                if (l_phase, j) in b.phase_component_set:
                    b.mole_frac_phase_comp[l_phase, j].value = b._mole_frac_tdew[
                        pp_VLE, j
                    ].value
        elif K is not None:
            _set_mole_fractions_vle(
                b,
                K,
                vap_frac,
                l_phase,
                v_phase,
                vl_comps + henry_mole_frac,
                l_only_comps,
                v_only_comps + henry_conc + henry_other,
            )
            if len(henry_conc) > 0:
                # With initial guesses for the mole fractions, there is a
                # decent value for density to calculate the concentration with
                for j in henry_conc:
                    K[j] = value(henry_equilibrium_ratio(b, l_phase, j))

                # Only recompute vapor fraction if we didn't get it from
                # dew and bubble point calculations
                if raoult_init:
                    vap_frac = _modified_rachford_rice(
                        b,
                        K,
                        vl_comps + henry_mole_frac + henry_conc,
                        l_only_comps,
                        v_only_comps + henry_other,
                    )
                    # If MRR failed with K[j] set for j in henry_conc, just
                    # leave the preliminary values assigned
                    if vap_frac is not None:
                        b.phase_frac[v_phase] = vap_frac
                        b.phase_frac[l_phase] = 1 - vap_frac

                        for p in pp_VLE:
                            b.flow_mol_phase[p].value = value(
                                b.phase_frac[p] * b.flow_mol
                            )
                        _set_mole_fractions_vle(
                            b,
                            K,
                            vap_frac,
                            l_phase,
                            v_phase,
                            vl_comps + henry_mole_frac + henry_conc,
                            l_only_comps,
                            v_only_comps + henry_other,
                        )
                else:
                    # Assign mole fractions for vap_frac from Rahul Gahndi's
                    # method
                    _set_mole_fractions_vle(
                        b,
                        K,
                        vap_frac,
                        l_phase,
                        v_phase,
                        vl_comps + henry_mole_frac + henry_conc,
                        l_only_comps,
                        v_only_comps + henry_other,
                    )


def define_default_scaling_factors(b):
    """
    Method to set default scaling factors for the property package. Scaling
    factors are based on the default initial value for each variable provided
    in the state_bounds config argument.
    """
    # Get bounds and initial values from config args
    units = b.get_metadata().derived_units
    state_bounds = b.config.state_bounds

    if state_bounds is None:
        return

    try:
        f_bounds = state_bounds["flow_mol"]
        if len(f_bounds) == 4:
            f_init = pyunits.convert_value(
                f_bounds[1], from_units=f_bounds[3], to_units=units.FLOW_MOLE
            )
        else:
            f_init = f_bounds[1]
    except KeyError:
        f_init = 1

    try:
        p_bounds = state_bounds["pressure"]
        if len(p_bounds) == 4:
            p_init = pyunits.convert_value(
                p_bounds[1], from_units=p_bounds[3], to_units=units.PRESSURE
            )
        else:
            p_init = p_bounds[1]
    except KeyError:
        p_init = 1

    try:
        t_bounds = state_bounds["temperature"]
        if len(t_bounds) == 4:
            t_init = pyunits.convert_value(
                t_bounds[1], from_units=t_bounds[3], to_units=units.TEMPERATURE
            )
        else:
            t_init = t_bounds[1]
    except KeyError:
        t_init = 1

    # Set default scaling factors
    b.set_default_scaling("flow_mol", 1 / f_init)
    b.set_default_scaling("flow_mol_phase", 1 / f_init)
    b.set_default_scaling("flow_mol_comp", 1 / f_init)
    b.set_default_scaling("flow_mol_phase_comp", 1 / f_init)
    b.set_default_scaling("pressure", 1 / p_init)
    b.set_default_scaling("temperature", 1 / t_init)


def calculate_scaling_factors(b):
    sf_flow = iscale.get_scaling_factor(b.flow_mol, default=1, warning=True)
    sf_mf = {}
    for i, v in b.mole_frac_phase_comp.items():
        sf_mf[i] = iscale.get_scaling_factor(v, default=1e3, warning=True)

    if b.config.defined_state is False:
        iscale.constraint_scaling_transform(
            b.sum_mole_frac_out, min(sf_mf.values()), overwrite=False
        )

    if len(b.phase_list) == 1:
        iscale.constraint_scaling_transform(
            b.total_flow_balance, sf_flow, overwrite=False
        )

        for j in b.component_list:
            sf_j = iscale.get_scaling_factor(
                b.mole_frac_comp[j], default=1e3, warning=True
            )
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_j, overwrite=False
            )

        # b.phase_fraction_constraint is well scaled

    elif len(b.phase_list) == 2:
        iscale.constraint_scaling_transform(
            b.total_flow_balance, sf_flow, overwrite=False
        )

        for j in b.component_list:
            sf_j = iscale.get_scaling_factor(
                b.mole_frac_comp[j], default=1e3, warning=True
            )
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_j * sf_flow, overwrite=False
            )

        iscale.constraint_scaling_transform(
            b.sum_mole_frac, min(sf_mf.values()), overwrite=False
        )

        for p in b.phase_list:
            iscale.constraint_scaling_transform(
                b.phase_fraction_constraint[p], sf_flow, overwrite=False
            )

    else:
        iscale.constraint_scaling_transform(
            b.total_flow_balance, sf_flow, overwrite=False
        )

        for j in b.component_list:
            sf_j = iscale.get_scaling_factor(
                b.mole_frac_comp[j], default=1e3, warning=True
            )
            iscale.constraint_scaling_transform(
                b.component_flow_balances[j], sf_j * sf_flow, overwrite=False
            )

        for p in b.phase_list:
            iscale.constraint_scaling_transform(
                b.sum_mole_frac[p], min(sf_mf[p, :].values()), overwrite=False
            )
            iscale.constraint_scaling_transform(
                b.phase_fraction_constraint[p], sf_flow, overwrite=False
            )

    if b.params._electrolyte:
        calculate_electrolyte_scaling(b)


do_not_initialize = ["sum_mole_frac_out"]


class FTPx(object):
    """Total flow, temperature, pressure, mole fraction state."""

    set_metadata = set_metadata
    define_state = define_state
    state_initialization = state_initialization
    do_not_initialize = do_not_initialize
    define_default_scaling_factors = define_default_scaling_factors
    calculate_scaling_factors = calculate_scaling_factors


def _set_mole_fractions_vle(
    b, K, vap_frac, l_phase, v_phase, vl_comps, l_only_comps, v_only_comps
):
    # The only reason this method exists is so we can use it twice because
    # concentration based Henry law coefficients make everything difficult
    for j in l_only_comps:
        b.mole_frac_phase_comp[l_phase, j].value = (
            1 / (1 - vap_frac) * b.mole_frac_comp[j]
        )

    for j in v_only_comps:
        b.mole_frac_phase_comp[v_phase, j].value = 1 / vap_frac * b.mole_frac_comp[j]

    for j in vl_comps:
        b.mole_frac_phase_comp[l_phase, j].value = value(
            b.mole_frac_comp[j] / (1 + vap_frac * (K[j] - 1))
        )
        b.mole_frac_phase_comp[v_phase, j].value = value(
            b.mole_frac_comp[j] * K[j] / (K[j] + 1 - vap_frac)
        )


def _modified_rachford_rice(b, K, vl_comps, l_only_comps, v_only_comps, eps=1e-5):
    """
    Uses a modified version of the Rachford Rice method to compute the vapor
    fraction for a two-phase vapor liquid equilibrium.

    Arguments:
        b: Property block for which this calculation is taking place
        K: Dictionary of equilibrium constants in the relationship
            y_j = K_j x_j for all j in vl_comps.
        vl_comps: List of components present in both phases
        l_only_comps: List of components present in only the liquid phase
        v_only_comps: List of components present in only the vapor phase
        eps (Optional): Offset from 0 or 1 in the event of a single phase
            solution

    Returns:
        phase_frac: Either a float between 0 and 1 if a root is found or
            RR predicts a single phase solution, or None if no root is found
            or some K_j < 0. Roots outside the range of 0 and 1 are clipped to
            be within that range.

    Notes:
        Mathematically, one can show that this method is guaranteed to give a
        meaningful answer given that:
            1. All K > 0 (any component with K_j = 0 should
                                            be designated an l_only_comp)
            2. All mole_frac_comp[j] > 0 in block b
            3. sum(mole_frac_comp[j] for j in b.component_list) == 1
        If the harmonic mean of K[j] is greater than one, the stream is pure
        vapor. If the arithmetic mean of K[j] is less than one, the stream is
        pure liquid. If neither of those conditions hold, a root exists between
        zero and one and the root-finding method will converge to it.

        However, validating user input to ensure that all these conditions
        hold turned out to be more trouble than it was worth (frequently users
        end up initializing with a sum of mole fractions less than one in
        reasonable use cases). Additionally, ill-conditioned problems open
        a gap between mathematical certainty and the results of floating point
        arithmetic.
    """
    # Validate split ratios K are nonnegative
    for j in vl_comps:
        if K[j] < 0:
            _log.warning(
                "While initializing block {}, the vapor/liquid split "
                "ratio of Component {} was calculated to be negative. "
                "Check the implementation of the saturation pressure, "
                "Henry's law method, or liquid density.".format(b.name, j)
            )
            return None

    # Calculate harmonic and arithmetic means of the split ratios KS
    if len(l_only_comps) == 0:
        tmp = sum([value(b.mole_frac_comp[j]) / K[j] for j in vl_comps])
        if tmp < 1e-14:
            K_bar_H = float("inf")
        else:
            K_bar_H = 1 / tmp
    else:
        K_bar_H = 0
    if len(v_only_comps) == 0:
        K_bar_A = sum([value(b.mole_frac_comp[j]) * K[j] for j in vl_comps])
    else:
        K_bar_A = float("inf")

    # If neither of these conditions hold, RR gives a single-phase solution
    if K_bar_H >= 1:
        # Vapor fraction is nearly one
        return 1 - eps
    elif K_bar_A <= 1:
        # Vapor fraction is nearly zero
        return eps

    # I discovered this method is a variant on solving the
    # Rachford-Rice equation, which has been known to ChemEs
    # since the 50s. I'm not sure if the link between Newton's
    # method and convexity was explicitly stated, though. For good
    # reason, the classic RR equation does not appear to necessarily
    # be convex

    # There are variants where the flash calculation is robust
    # to the calculated vapor fraction being greater than 1 or
    # less than zero. Future reader, check out
    # https://doi.org/10.1016/j.fluid.2011.12.005 and
    # https://doi.org/10.1016/S0378-3812(97)00179-9 if it
    # becomes necessary to increase robustness to very nonideal
    # systems.

    # Equation defining vapor fraction for an ideal mixture.
    # This equation has two roots: a trivial one at vap_frac = 1
    # and a nontrivial one for some 0<vap_frac<1. B is a
    # convex function
    def B(vap_frac):
        return (
            sum(
                [
                    value(b.mole_frac_comp[j] * K[j] / (1 + vap_frac * (K[j] - 1)))
                    for j in vl_comps
                ]
            )
            + sum([value(b.mole_frac_comp[j] / vap_frac) for j in v_only_comps])
            - 1
        )

    def dB_dvap_frac(vap_frac):
        return sum(
            [
                value(
                    -b.mole_frac_comp[j]
                    * K[j]
                    * (K[j] - 1)
                    / ((1 + vap_frac * (K[j] - 1)) ** 2)
                )
                for j in vl_comps
            ]
        ) - sum([value(b.mole_frac_comp[j] / vap_frac**2) for j in v_only_comps])

    # Newton's method will always undershoot the root of a
    # convex equation if one starts from a value from which
    # the function is positive. However, there is a singularity
    # at vap_frac=0 if there are noncondensable components.
    # Therefore, use a binary search to find a value where
    # B(vap_frac) is positive
    k = 0
    vap_frac = 0.5
    mRR_failed = False
    while B(vap_frac) < 0:
        vap_frac /= 2
        k += 1
        if k > 40:
            mRR_failed = True
            break
    # Now use Newton's method to calculate vap_frac
    k = 0
    if not mRR_failed:
        while abs(B(vap_frac)) > 1e-6:
            vap_frac -= B(vap_frac) / dB_dvap_frac(vap_frac)
            k += 1
            if k > 40:
                mRR_failed = True
                break
    if mRR_failed:
        # mRR did not even converge to a root. Do not attempt to assign
        # a vapor fraction
        vap_frac = None
    elif vap_frac < 0:
        # mRR converged to a negative root. Initialize as mostly liquid
        vap_frac = eps
        mRR_failed = True
    elif vap_frac > 1:
        # mRR converged to a root > 1. Initialize as mostly vapor
        vap_frac = 1 - eps
        mRR_failed = True
    # Regardless of the failure mode, log a warning
    if mRR_failed:
        _log.warning(
            "Block {} - phase faction initialization using "
            "modified Rachford-Rice failed. This could be "
            "because a component is essentially "
            "nonvolatile or noncondensible, or "
            "because mole fractions sum to more than "
            "one.".format(b.name)
        )
    return vap_frac



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example Peng-Robinson parameter block for the VLE calucations for a
benzene-toluene system.

Unless otherwise noted, parameters are from:
"The Properties of Gases and Liquids, 4th Edition", Reid, Prausnitz and Poling,
McGraw-Hill, 1987
"""
# Import Pyomo libraries
from pyomo.environ import Reals, Param, NonNegativeReals, Set, units as pyunits

# Import IDAES cores
from idaes.core import declare_process_block_class, Component
from idaes.core.util.misc import extract_data

from idaes.models.properties.cubic_eos.cubic_prop_pack import (
    CubicParameterData,
    CubicEoS,
)
from idaes.logger import getIdaesLogger


# Set up logger
_log = getIdaesLogger(__name__)


@declare_process_block_class("BTParameterBlock")
class BTParameterData(CubicParameterData):
    """Cubic property package for benzene and toluene mixtures."""

    def build(self):
        """
        Callable method for Block construction.
        """
        super(BTParameterData, self).build()

        self.cubic_type = CubicEoS.PR

        # Add Component objects
        self.benzene = Component(elemental_composition={"C": 6, "H": 6})
        self.toluene = Component(elemental_composition={"C": 7, "H": 8})

        # List of phase equilibrium index
        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        self.phase_equilibrium_list = {
            1: ["benzene", ("Vap", "Liq")],
            2: ["toluene", ("Vap", "Liq")],
        }

        # Thermodynamic reference state
        self.pressure_ref = Param(
            mutable=True,
            default=101325,
            doc="Reference pressure [Pa]",
            units=pyunits.Pa,
        )
        self.temperature_ref = Param(
            mutable=True,
            default=298.15,
            doc="Reference temperature [K]",
            units=pyunits.K,
        )

        # Critical Properties
        pressure_crit_data = {"benzene": 48.9e5, "toluene": 41.0e5}

        self.pressure_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(pressure_crit_data),
            doc="Critical pressure [Pa]",
            units=pyunits.Pa,
        )

        temperature_crit_data = {"benzene": 562.2, "toluene": 591.8}

        self.temperature_crit = Param(
            self.component_list,
            within=NonNegativeReals,
            mutable=False,
            initialize=extract_data(temperature_crit_data),
            doc="Critical temperature [K]",
            units=pyunits.K,
        )

        # Pitzer acentricity factor
        omega_data = {"benzene": 0.212, "toluene": 0.263}

        self.omega = Param(
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(omega_data),
            doc="Acentricity Factor",
        )

        # Peng-Robinson binary interaction parameters
        kappa_data = {
            ("benzene", "benzene"): 0.0000,
            ("benzene", "toluene"): 0.0000,
            ("toluene", "benzene"): 0.0000,
            ("toluene", "toluene"): 0.0000,
        }

        self.kappa = Param(
            self.component_list,
            self.component_list,
            within=Reals,
            mutable=False,
            initialize=extract_data(kappa_data),
            doc="Peng-Robinson binary interaction parameters",
        )

        # Molecular Weights
        mw_comp_data = {"benzene": 78.1136e-3, "toluene": 92.1405e-3}

        self.mw_comp = Param(
            self.component_list,
            mutable=False,
            initialize=extract_data(mw_comp_data),
            doc="molecular weight kg/mol",
            units=pyunits.kg / pyunits.mol,
        )

        # Constants for specific heat capacity, enthalpy and entropy
        self.cp_mol_ig_comp_coeff_1 = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": -3.392e1, "toluene": -2.435e1},
            doc="Parameter 1 to compute cp_mol_comp",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )

        self.cp_mol_ig_comp_coeff_2 = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": 4.739e-1, "toluene": 5.125e-1},
            doc="Parameter 2 to compute cp_mol_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**2,
        )

        self.cp_mol_ig_comp_coeff_3 = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": -3.017e-4, "toluene": -2.765e-4},
            doc="Parameter 3 to compute cp_mol_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**3,
        )

        self.cp_mol_ig_comp_coeff_4 = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": 7.130e-8, "toluene": 4.911e-8},
            doc="Parameter 4 to compute cp_mol_comp",
            units=pyunits.J / pyunits.mol / pyunits.K**4,
        )

        # Standard heats of formation
        # Source: NIST Webbook, https://webbook.nist.gov
        # Retrieved 25th September 2019
        dh_form_data = {"benzene": 82.9e3, "toluene": 50.1e3}

        self.enth_mol_form_ref = Param(
            self.component_list,
            mutable=False,
            initialize=extract_data(dh_form_data),
            doc="Standard heats of formation",
            units=pyunits.J / pyunits.mol,
        )

        # Standard entropy of formation
        # Source: Engineering Toolbox, https://www.engineeringtoolbox.com
        # Retrieved 25th September, 2019
        ds_form_data = {"benzene": -269, "toluene": -321}

        self.entr_mol_form_ref = Param(
            self.component_list,
            mutable=False,
            initialize=extract_data(ds_form_data),
            doc="Standard entropy of formation",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )

        # Antoine coefficients for ideal vapour (units: bar, K)
        # This is needed for initial guesses of bubble and dew points
        self.antoine_coeff_A = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": 4.202, "toluene": 4.216},
            doc="Antoine A Parameter to calculate pressure_sat",
            units=pyunits.dimensionless,
        )

        self.antoine_coeff_B = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": 1322, "toluene": 1435},
            doc="Antoine B Parameter to calculate pressure_sat",
            units=pyunits.K,
        )

        self.antoine_coeff_C = Param(
            self.component_list,
            mutable=False,
            initialize={"benzene": -38.56, "toluene": -43.33},
            doc="Antoine C Parameter to calculate pressure_sat",
            units=pyunits.K,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions for converting units of measurement and
reporting model outputs.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import as_quantity, units
import idaes


def report_quantity(c):
    q = as_quantity(c)

    return convert_quantity_to_reporting_units(q)


def convert_quantity_to_reporting_units(q):
    """
    Converts a pint quantity to the units defined in the IDAES config block.

    Args:
        q: pint quantity to be converted

    Returns:
        A new pint quantity in the units defined by the IDAES config block.
    """
    # First, check if quantity is dimensionless
    if q.dimensionless:
        # No need to do anything here
        return q

    # Get definition of desired units from config block
    def_units = idaes.cfg.reporting_units

    # Get dimensionality of q
    dim = q.dimensionality

    # Iterate through unit definition to try to find matching dimensionality
    for u in def_units.values():
        # Get pint unit object from string
        u_obj = getattr(units.pint_registry, u)
        if str(u_obj.dimensionality) == str(dim):
            # Found matching dimensionality
            q2 = q.to(u_obj)
            break
    else:
        # No matching dimensionality found, fall back to default system of units
        q2 = q.to_base_units()

    return q2



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating equilibrium constants
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from pyomo.environ import exp, log, Var, units as pyunits, value

from idaes.models.properties.modular_properties.base.utility import ConcentrationForm
from idaes.core.util.misc import set_param_from_config
from idaes.core.util.constants import Constants as c
from idaes.core.util.exceptions import BurntToast, ConfigurationError
from .dh_rxn import constant_dh_rxn


# -----------------------------------------------------------------------------
# Constant Keq
class ConstantKeq:
    """Methods for invariant equilibrium constant."""

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name)
            )
        elif (
            c_form == ConcentrationForm.moleFraction
            or c_form == ConcentrationForm.massFraction
            or c_form == ConcentrationForm.activity
        ):
            e_units = pyunits.dimensionless
        else:
            order = 0

            try:
                # This will work for Reaction Packages
                pc_set = parent.config.property_package._phase_component_set
            except AttributeError:
                # Need to allow for inherent reactions in Property Packages
                if not parent._electrolyte:
                    # In most cases ,should have _phase_component_set
                    pc_set = parent._phase_component_set
                else:
                    # However, for electrolytes need true species set
                    pc_set = parent.true_phase_component_set

            for p, j in pc_set:
                order += rblock.reaction_order[p, j].value

            if c_form == ConcentrationForm.molarity:
                c_units = units.DENSITY_MOLE
            elif c_form == ConcentrationForm.molality:
                c_units = units.AMOUNT * units.MASS**-1
            elif c_form == ConcentrationForm.partialPressure:
                c_units = units.PRESSURE
            else:
                raise BurntToast(
                    "{} received unrecognised ConcentrationForm ({}). "
                    "This should not happen - please contact the IDAES "
                    "developers with this bug.".format(rblock.name, c_form)
                )

            e_units = c_units**order

        rblock.k_eq_ref = Var(
            doc="Equilibrium constant at reference state", units=e_units
        )
        set_param_from_config(rblock, param="k_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        return rblock.k_eq_ref

    @staticmethod
    def return_log_expression(b, rblock, r_idx, T):
        units = pyunits.get_units(rblock.k_eq_ref)
        if units is None or units == pyunits.dimensionless:
            expr = rblock.k_eq_ref
        else:
            expr = rblock.k_eq_ref / units
        return b.log_k_eq[r_idx] == log(expr)

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        return 1 / value(rblock.k_eq_ref)


# -----------------------------------------------------------------------------
# van t'Hoff equation (constant dh_rxn)
class van_t_hoff:
    """Methods for equilibrium constant using van t'Hoff equation"""

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name)
            )
        elif (
            c_form == ConcentrationForm.moleFraction
            or c_form == ConcentrationForm.massFraction
            or c_form == ConcentrationForm.activity
        ):
            e_units = pyunits.dimensionless
        else:
            order = 0

            try:
                # This will work for Reaction Packages
                pc_set = parent.config.property_package._phase_component_set
            except AttributeError:
                # Need to allow for inherent reactions in Property Packages
                if not parent._electrolyte:
                    # In most cases ,should have _phase_component_set
                    pc_set = parent._phase_component_set
                else:
                    # However, for electrolytes need true species set
                    pc_set = parent.true_phase_component_set

            for p, j in pc_set:
                order += rblock.reaction_order[p, j].value

            if c_form == ConcentrationForm.molarity:
                c_units = units.DENSITY_MOLE
            elif c_form == ConcentrationForm.molality:
                c_units = units.AMOUNT * units.MASS**-1
            elif c_form == ConcentrationForm.partialPressure:
                c_units = units.PRESSURE
            else:
                raise BurntToast(
                    "{} received unrecognised ConcentrationForm ({}). "
                    "This should not happen - please contact the IDAES "
                    "developers with this bug.".format(rblock.name, c_form)
                )

            e_units = c_units**order

        rblock.k_eq_ref = Var(
            doc="Equilibrium constant at reference state", units=e_units
        )
        set_param_from_config(rblock, param="k_eq_ref", config=config)

        rblock.T_eq_ref = Var(
            doc="Reference temperature for equilibrium constant",
            units=units.TEMPERATURE,
        )
        set_param_from_config(rblock, param="T_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        units = pyunits.get_units(rblock.k_eq_ref)
        if units is None or units is pyunits.dimensionless:
            return exp(b.log_k_eq[r_idx])
        else:
            return exp(b.log_k_eq[r_idx]) * units

    @staticmethod
    def return_log_expression(b, rblock, r_idx, T):
        units = rblock.parent_block().get_metadata().derived_units

        k_units = pyunits.get_units(rblock.k_eq_ref)

        if k_units is None or k_units is pyunits.dimensionless:
            l_keq_ref = log(rblock.k_eq_ref)
        else:
            l_keq_ref = log(rblock.k_eq_ref / k_units)

        return (b.log_k_eq[r_idx] - l_keq_ref) == (
            -b.dh_rxn[r_idx]
            / pyunits.convert(c.gas_constant, to_units=units.GAS_CONSTANT)
            * (1 / T - 1 / rblock.T_eq_ref)
        )

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        return 1 / value(rblock.k_eq_ref)


# -----------------------------------------------------------------------------
# Constant dh_rxn and ds_rxn
class gibbs_energy:
    """Methods for equilibrium constant based of constant heat and entropy of reaction."""

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name)
            )
        elif not (
            c_form == ConcentrationForm.moleFraction
            or c_form == ConcentrationForm.activity
        ):
            raise ConfigurationError(
                "{} calculation of equilibrium constant based on Gibbs energy "
                "is only supported for mole fraction or activity forms. "
                "Currently selected form: {}".format(rblock.name, c_form)
            )

        # Check that heat of reaction is constant
        if config.heat_of_reaction is not constant_dh_rxn:
            raise ConfigurationError(
                "{} calculating equilibrium constants from Gibbs energy "
                "assumes constant heat of reaction. Please ensure you are "
                "using the constant_dh_rxn method for this reaction".format(rblock.name)
            )

        rblock.ds_rxn_ref = Var(
            doc="Specific molar entropy of reaction at reference state",
            units=units.ENTROPY_MOLE,
        )
        set_param_from_config(rblock, param="ds_rxn_ref", config=config)

        rblock.T_eq_ref = Var(
            doc="Reference temperature for equilibrium constant",
            units=units.TEMPERATURE,
        )
        set_param_from_config(rblock, param="T_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        return exp(b.log_k_eq[r_idx])

    @staticmethod
    def return_log_expression(b, rblock, r_idx, T):
        units = rblock.parent_block().get_metadata().derived_units
        R = pyunits.convert(c.gas_constant, to_units=units.GAS_CONSTANT)

        return b.log_k_eq[r_idx] == (
            (-rblock.dh_rxn_ref / (R * T)) + (rblock.ds_rxn_ref / R)
        )

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        units = rblock.parent_block().get_metadata().derived_units
        R = pyunits.convert(c.gas_constant, to_units=units.GAS_CONSTANT)

        keq_val = value(
            exp(-rblock.dh_rxn_ref / (R * rblock.T_eq_ref) + rblock.ds_rxn_ref / R)
            * rblock._keq_units
        )
        return 1 / keq_val



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Plate Heat Exchnager  unit model.
Author: Akula Paul
"""
import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Param,
    units as pyunits,
    value,
)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core import FlowsheetBlock
from idaes.models_extra.column_models.plate_heat_exchanger import (
    PlateHeatExchanger as PHE,
)

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.column_models.properties.MEA_solvent import (
    configuration as aqueous_mea,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.hotside_properties = GenericParameterBlock(**aqueous_mea)
    m.fs.coldside_properties = GenericParameterBlock(**aqueous_mea)

    m.fs.unit = PHE(
        passes=4,
        channels_per_pass=12,
        number_of_divider_plates=2,
        hot_side={"property_package": m.fs.hotside_properties},
        cold_side={"property_package": m.fs.coldside_properties},
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 9


workaround_for_1294 = pytest.mark.xfail(
    # the failures only occur for Windows on GHA with Python <3.12, and Linux with Python 3.12
    reason="These tests fail with Pyomo 6.7.0. See IDAES/idaes-pse#1294 for details",
    strict=False,
)


# -----------------------------------------------------------------------------
class TestPHE(object):
    @pytest.fixture(scope="class")
    def phe(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.hotside_properties = GenericParameterBlock(**aqueous_mea)
        m.fs.coldside_properties = GenericParameterBlock(**aqueous_mea)

        m.fs.unit = PHE(
            passes=4,
            channels_per_pass=12,
            number_of_divider_plates=2,
            hot_side={"property_package": m.fs.hotside_properties},
            cold_side={"property_package": m.fs.coldside_properties},
        )

        # hot fluid
        m.fs.unit.hot_side_inlet.flow_mol[0].fix(60.54879)
        m.fs.unit.hot_side_inlet.temperature[0].fix(392.23)
        m.fs.unit.hot_side_inlet.pressure[0].fix(202650)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "CO2"].fix(0.0158)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "H2O"].fix(0.8747)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "MEA"].fix(0.1095)

        # cold fluid
        m.fs.unit.cold_side_inlet.flow_mol[0].fix(63.01910)
        m.fs.unit.cold_side_inlet.temperature[0].fix(326.36)
        m.fs.unit.cold_side_inlet.pressure[0].fix(202650)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "CO2"].fix(0.0414)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "H2O"].fix(0.8509)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "MEA"].fix(0.1077)

        # Fix unit geometry - default values should be correct
        m.fs.unit.plate_length.fix()
        m.fs.unit.plate_width.fix()
        m.fs.unit.plate_thickness.fix()
        m.fs.unit.plate_pact_length.fix()
        m.fs.unit.port_diameter.fix()
        m.fs.unit.plate_therm_cond.fix()
        m.fs.unit.area.fix()

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, phe):

        assert hasattr(phe.fs.unit, "hot_side_inlet")
        assert len(phe.fs.unit.hot_side_inlet.vars) == 4
        assert hasattr(phe.fs.unit.hot_side_inlet, "flow_mol")
        assert hasattr(phe.fs.unit.hot_side_inlet, "mole_frac_comp")
        assert hasattr(phe.fs.unit.hot_side_inlet, "temperature")
        assert hasattr(phe.fs.unit.hot_side_inlet, "pressure")

        assert hasattr(phe.fs.unit, "hot_side_outlet")
        assert len(phe.fs.unit.hot_side_outlet.vars) == 4
        assert hasattr(phe.fs.unit.hot_side_outlet, "flow_mol")
        assert hasattr(phe.fs.unit.hot_side_outlet, "mole_frac_comp")
        assert hasattr(phe.fs.unit.hot_side_outlet, "temperature")
        assert hasattr(phe.fs.unit.hot_side_outlet, "pressure")

        assert hasattr(phe.fs.unit, "cold_side_inlet")
        assert len(phe.fs.unit.cold_side_inlet.vars) == 4
        assert hasattr(phe.fs.unit.cold_side_inlet, "flow_mol")
        assert hasattr(phe.fs.unit.cold_side_inlet, "mole_frac_comp")
        assert hasattr(phe.fs.unit.cold_side_inlet, "temperature")
        assert hasattr(phe.fs.unit.cold_side_inlet, "pressure")

        assert hasattr(phe.fs.unit, "cold_side_outlet")
        assert len(phe.fs.unit.cold_side_outlet.vars) == 4
        assert hasattr(phe.fs.unit.cold_side_outlet, "flow_mol")
        assert hasattr(phe.fs.unit.cold_side_outlet, "mole_frac_comp")
        assert hasattr(phe.fs.unit.cold_side_outlet, "temperature")
        assert hasattr(phe.fs.unit.cold_side_outlet, "pressure")

        assert hasattr(phe.fs.unit.cold_side, "deltaP")
        assert hasattr(phe.fs.unit.hot_side, "deltaP")

        assert isinstance(phe.fs.unit.number_of_passes, Param)
        assert isinstance(phe.fs.unit.channels_per_pass, Param)

    @pytest.mark.component
    def test_units(self, phe):
        assert_units_equivalent(phe.fs.unit.Re_hot[0], pyunits.dimensionless)
        assert_units_equivalent(phe.fs.unit.Re_cold[0], pyunits.dimensionless)
        assert_units_consistent(phe)

    @pytest.mark.unit
    def test_dof(self, phe):
        assert degrees_of_freedom(phe) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, phe):
        initialization_tester(
            phe, duty=(245000, pyunits.W), optarg={"bound_push": 1e-8, "mu_init": 1e-8}
        )

    @workaround_for_1294
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, phe):
        results = solver.solve(phe)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @workaround_for_1294
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, phe):
        # phe.fs.unit.display()
        assert pytest.approx(182244.65, rel=1e-5) == value(
            phe.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(177366.53, rel=1e-5) == value(
            phe.fs.unit.cold_side_outlet.pressure[0]
        )

        assert pytest.approx(685.730, rel=1e-5) == value(phe.fs.unit.Re_hot[0])
        assert pytest.approx(196.018, rel=1e-5) == value(phe.fs.unit.Re_cold[0])
        assert pytest.approx(4.55441, rel=1e-5) == value(phe.fs.unit.Pr_hot[0])
        assert pytest.approx(15.0566, rel=1e-5) == value(phe.fs.unit.Pr_cold[0])
        assert pytest.approx(3691.35, rel=1e-5) == value(
            phe.fs.unit.heat_transfer_coefficient_hot_side[0]
        )
        assert pytest.approx(2610.81, rel=1e-5) == value(
            phe.fs.unit.heat_transfer_coefficient_cold_side[0]
        )
        assert pytest.approx(1170.34, rel=1e-5) == value(
            phe.fs.unit.heat_transfer_coefficient[0]
        )

        assert pytest.approx(24.4118, rel=1e-5) == value(phe.fs.unit.NTU[0])
        assert pytest.approx(0.971227, rel=1e-5) == value(phe.fs.unit.Cratio[0])
        assert pytest.approx(0.676902, rel=1e-5) == value(phe.fs.unit.effectiveness[0])

        assert pytest.approx(244327, rel=1e-5) == value(phe.fs.unit.heat_duty[0])

        assert pytest.approx(345.612, rel=1e-5) == value(
            phe.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(372.776, rel=1e-5) == value(
            phe.fs.unit.cold_side_outlet.temperature[0]
        )

    @workaround_for_1294
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, phe):
        # Mass conservation test
        assert (
            abs(
                value(
                    phe.fs.unit.hot_side_inlet.flow_mol[0]
                    - phe.fs.unit.hot_side_outlet.flow_mol[0]
                )
            )
            <= 2e-6
        )

        for j in phe.fs.hotside_properties.apparent_species_set:
            assert (
                abs(
                    value(
                        phe.fs.unit.hot_side_inlet.flow_mol[0]
                        * phe.fs.unit.hot_side_inlet.mole_frac_comp[0, j]
                        - phe.fs.unit.hot_side_outlet.flow_mol[0]
                        * phe.fs.unit.hot_side_outlet.mole_frac_comp[0, j]
                    )
                )
                <= 2e-6
            )

        assert (
            abs(
                value(
                    phe.fs.unit.cold_side_inlet.flow_mol[0]
                    - phe.fs.unit.cold_side_outlet.flow_mol[0]
                )
            )
            <= 2e-6
        )

        for j in phe.fs.coldside_properties.apparent_species_set:
            assert (
                abs(
                    value(
                        phe.fs.unit.cold_side_inlet.flow_mol[0]
                        * phe.fs.unit.cold_side_inlet.mole_frac_comp[0, j]
                        - phe.fs.unit.cold_side_outlet.flow_mol[0]
                        * phe.fs.unit.cold_side_outlet.mole_frac_comp[0, j]
                    )
                )
                <= 2e-6
            )

        # Energy conservation test
        assert (
            abs(
                value(
                    phe.fs.unit.hot_side.properties_in[0]._enthalpy_flow_term["Liq"]
                    + phe.fs.unit.cold_side.properties_in[0]._enthalpy_flow_term["Liq"]
                    - phe.fs.unit.hot_side.properties_out[0]._enthalpy_flow_term["Liq"]
                    - phe.fs.unit.cold_side.properties_out[0]._enthalpy_flow_term["Liq"]
                )
            )
            <= 2e-6
        )

    # @pytest.mark.ui
    # @pytest.mark.unit
    # def test_report(self, phe):
    #     phe.fs.unit.report()
